# Copyright (c) OpenMMLab. All rights reserved.
from copy import deepcopy

import numpy as np
import torch
import torch.nn as nn
from mmengine.logging import MMLogger
from mmengine.model import xavier_init
from mmengine.runner import load_checkpoint
from mmengine.runner.checkpoint import _load_checkpoint_with_prefix
from torch.nn.init import xavier_uniform_
from torch.nn.utils import spectral_norm

from mmagic.registry import MODELS


@MODELS.register_module('SAGANDiscriminator')
@MODELS.register_module()
class ProjDiscriminator(nn.Module):
    r"""Discriminator for SNGAN / Proj-GAN. The implementation is refer to
    https://github.com/pfnet-research/sngan_projection/tree/master/dis_models

    The overall structure of the projection discriminator can be split into a
    ``from_rgb`` layer, a group of ResBlocks, a linear decision layer, and a
    projection layer. To support defining custom layers, we introduce
    ``from_rgb_cfg`` and ``blocks_cfg``.

    The design of the model structure is highly corresponding to the output
    resolution. Therefore, we provide `channels_cfg` and `downsample_cfg` to
    control the input channels and the downsample behavior of the intermediate
    blocks.

    ``downsample_cfg``: In default config of SNGAN / Proj-GAN, whether to apply
        downsample in each intermediate blocks is quite flexible and
        corresponding to the resolution of the output image. Therefore, we
        support user to define the ``downsample_cfg`` by themselves, and to
        control the structure of the discriminator.

    ``channels_cfg``: In default config of SNGAN / Proj-GAN, the number of
        ResBlocks and the channels of those blocks are corresponding to the
        resolution of the output image. Therefore, we allow user to define
        `channels_cfg` for try their own models.  We also provide a default
        config to allow users to build the model only from the output
        resolution.

    Args:
        input_scale (int): The scale of the input image.
        num_classes (int, optional): The number classes you would like to
            generate. If num_classes=0, no label projection would be used.
            Default to 0.
        base_channels (int, optional): The basic channel number of the
            discriminator. The other layers contains channels based on this
            number.  Defaults to 128.
        input_channels (int, optional): Channels of the input image.
            Defaults to 3.
        attention_cfg (dict, optional): Config for the self-attention block.
            Default to ``dict(type='SelfAttentionBlock')``.
        attention_after_nth_block (int | list[int], optional): Self-attention
            block would be added after which *ConvBlock* (including the head
            block). If ``int`` is passed, only one attention block would be
            added. If ``list`` is passed, self-attention blocks would be added
            after multiple ConvBlocks. To be noted that if the input is
            smaller than ``1``, self-attention corresponding to this index
            would be ignored. Default to 0.
        channels_cfg (list | dict[list], optional): Config for input channels
            of the intermediate blocks. If list is passed, each element of the
            list means the input channels of current block is how many times
            compared to the ``base_channels``. For block ``i``, the input and
            output channels should be ``channels_cfg[i]`` and
            ``channels_cfg[i+1]`` If dict is provided, the key of the dict
            should be the output scale and corresponding value should be a list
            to define channels.  Default: Please refer to
            ``_defualt_channels_cfg``.
        downsample_cfg (list[bool] | dict[list], optional): Config for
            downsample behavior of the intermediate layers. If a list is
            passed, ``downsample_cfg[idx] == True`` means apply downsample in
            idx-th block, and vice versa. If dict is provided, the key dict
            should be the input scale of the image and corresponding value
            should be a list ti define the downsample behavior. Default: Please
            refer to ``_default_downsample_cfg``.
        from_rgb_cfg (dict, optional): Config for the first layer to convert
            rgb image to feature map. Defaults to
            ``dict(type='SNGANDiscHeadResBlock')``.
        blocks_cfg (dict, optional): Config for the intermediate blocks.
            Defaults to ``dict(type='SNGANDiscResBlock')``
        act_cfg (dict, optional): Activation config for the final output
            layer. Defaults to ``dict(type='ReLU')``.
        with_spectral_norm (bool, optional): Whether use spectral norm for
            all conv blocks or not. Default to True.
        sn_style (str, optional): The style of spectral normalization.
            If set to `ajbrock`, implementation by
            ajbrock(https://github.com/ajbrock/BigGAN-PyTorch/blob/master/layers.py)
            will be adopted.
            If set to `torch`, implementation by `PyTorch` will be adopted.
            Defaults to `torch`.
        sn_eps (float, optional): eps for spectral normalization operation.
            Defaults to `1e-12`.
        init_cfg (dict, optional): Config for weight initialization.
            Default to ``dict(type='BigGAN')``.
        pretrained (str | dict , optional): Path for the pretrained model or
            dict containing information for pretrained models whose necessary
            key is 'ckpt_path'. Besides, you can also provide 'prefix' to load
            the generator part from the whole state dict.  Defaults to None.
    """

    # default channel factors
    _defualt_channels_cfg = {
        32: [1, 1, 1],
        64: [2, 4, 8, 16],
        128: [2, 4, 8, 16, 16],
    }

    # default downsample behavior
    _defualt_downsample_cfg = {
        32: [True, False, False],
        64: [True, True, True, True],
        128: [True, True, True, True, False]
    }

    def __init__(self,
                 input_scale,
                 num_classes=0,
                 base_channels=128,
                 input_channels=3,
                 attention_cfg=dict(type='SelfAttentionBlock'),
                 attention_after_nth_block=-1,
                 channels_cfg=None,
                 downsample_cfg=None,
                 from_rgb_cfg=dict(type='SNGANDiscHeadResBlock'),
                 blocks_cfg=dict(type='SNGANDiscResBlock'),
                 act_cfg=dict(type='ReLU'),
                 with_spectral_norm=True,
                 sn_style='torch',
                 sn_eps=1e-12,
                 init_cfg=dict(type='BigGAN'),
                 pretrained=None):

        super().__init__()

        self.init_type = init_cfg.get('type', None)

        # add SN options and activation function options to cfg
        self.from_rgb_cfg = deepcopy(from_rgb_cfg)
        self.from_rgb_cfg.setdefault('act_cfg', act_cfg)
        self.from_rgb_cfg.setdefault('with_spectral_norm', with_spectral_norm)
        self.from_rgb_cfg.setdefault('sn_style', sn_style)
        self.from_rgb_cfg.setdefault('init_cfg', init_cfg)

        # add SN options and activation function options to cfg
        self.blocks_cfg = deepcopy(blocks_cfg)
        self.blocks_cfg.setdefault('act_cfg', act_cfg)
        self.blocks_cfg.setdefault('with_spectral_norm', with_spectral_norm)
        self.blocks_cfg.setdefault('sn_style', sn_style)
        self.blocks_cfg.setdefault('sn_eps', sn_eps)
        self.blocks_cfg.setdefault('init_cfg', init_cfg)

        channels_cfg = deepcopy(self._defualt_channels_cfg) \
            if channels_cfg is None else deepcopy(channels_cfg)
        if isinstance(channels_cfg, dict):
            if input_scale not in channels_cfg:
                raise KeyError(f'`input_scale={input_scale} is not found in '
                               '`channel_cfg`, only support configs for '
                               f'{[chn for chn in channels_cfg.keys()]}')
            self.channel_factor_list = channels_cfg[input_scale]
        elif isinstance(channels_cfg, list):
            self.channel_factor_list = channels_cfg
        else:
            raise ValueError('Only support list or dict for `channel_cfg`, '
                             f'receive {type(channels_cfg)}')

        downsample_cfg = deepcopy(self._defualt_downsample_cfg) \
            if downsample_cfg is None else deepcopy(downsample_cfg)
        if isinstance(downsample_cfg, dict):
            if input_scale not in downsample_cfg:
                raise KeyError(f'`output_scale={input_scale} is not found in '
                               '`downsample_cfg`, only support configs for '
                               f'{[chn for chn in downsample_cfg.keys()]}')
            self.downsample_list = downsample_cfg[input_scale]
        elif isinstance(downsample_cfg, list):
            self.downsample_list = downsample_cfg
        else:
            raise ValueError('Only support list or dict for `channel_cfg`, '
                             f'receive {type(downsample_cfg)}')

        if len(self.downsample_list) != len(self.channel_factor_list):
            raise ValueError('`downsample_cfg` should have same length with '
                             '`channels_cfg`, but receive '
                             f'{len(self.downsample_list)} and '
                             f'{len(self.channel_factor_list)}.')

        # check `attention_after_nth_block`
        if not isinstance(attention_after_nth_block, list):
            attention_after_nth_block = [attention_after_nth_block]
        if not all([isinstance(idx, int)
                    for idx in attention_after_nth_block]):
            raise ValueError('`attention_after_nth_block` only support int or '
                             'a list of int. Please check your input type.')

        self.from_rgb = MODELS.build(
            self.from_rgb_cfg,
            default_args=dict(
                in_channels=input_channels, out_channels=base_channels))

        self.conv_blocks = nn.ModuleList()
        # add self-attention block after the first block
        if 1 in attention_after_nth_block:
            attn_cfg_ = deepcopy(attention_cfg)
            attn_cfg_['in_channels'] = base_channels
            attn_cfg_['sn_style'] = sn_style
            self.conv_blocks.append(MODELS.build(attn_cfg_))

        for idx in range(len(self.downsample_list)):
            factor_input = 1 if idx == 0 else self.channel_factor_list[idx - 1]
            factor_output = self.channel_factor_list[idx]

            # get block-specific config
            block_cfg_ = deepcopy(self.blocks_cfg)
            block_cfg_['downsample'] = self.downsample_list[idx]
            block_cfg_['in_channels'] = factor_input * base_channels
            block_cfg_['out_channels'] = factor_output * base_channels
            self.conv_blocks.append(MODELS.build(block_cfg_))

            # build self-attention block
            # the first ConvBlock is `from_rgb` block,
            # add 2 to get the index of the ConvBlocks
            if idx + 2 in attention_after_nth_block:
                attn_cfg_ = deepcopy(attention_cfg)
                attn_cfg_['in_channels'] = factor_output * base_channels
                self.conv_blocks.append(MODELS.build(attn_cfg_))

        self.decision = nn.Linear(factor_output * base_channels, 1)

        if with_spectral_norm:
            self.decision = spectral_norm(self.decision)

        self.num_classes = num_classes

        # In this case, discriminator is designed for conditional synthesis.
        if num_classes > 0:
            self.proj_y = nn.Embedding(num_classes,
                                       factor_output * base_channels)
            if with_spectral_norm:
                self.proj_y = spectral_norm(self.proj_y)

        self.activate = MODELS.build(act_cfg)
        self.init_weights(pretrained)

    def forward(self, x, label=None):
        """Forward function. If `self.num_classes` is larger than 0, label
        projection would be used.

        Args:
            x (torch.Tensor): Fake or real image tensor.
            label (torch.Tensor, options): Label correspond to the input image.
                Noted that, if `self.num_classed` is larger than 0,
                `label` should not be None.  Default to None.

        Returns:
            torch.Tensor: Prediction for the reality of the input image.
        """
        h = self.from_rgb(x)
        for conv_block in self.conv_blocks:
            h = conv_block(h)
        h = self.activate(h)
        h = torch.sum(h, dim=[2, 3])
        out = self.decision(h)

        if self.num_classes > 0:
            w_y = self.proj_y(label)
            out = out + torch.sum(w_y * h, dim=1, keepdim=True)
        return out.view(out.size(0), -1)

    def init_weights(self, pretrained=None, strict=True):
        """Init weights for SNGAN-Proj and SAGAN. If ``pretrained=None`` and
        weight initialization would follow the ``INIT_TYPE`` in
        ``init_cfg=dict(type=INIT_TYPE)``.

        For SNGAN-Proj
        (``INIT_TYPE.upper() in ['SNGAN', 'SNGAN-PROJ', 'GAN-PROJ']``),
        we follow the initialization method in the official Chainer's
        implementation (https://github.com/pfnet-research/sngan_projection).

        For SAGAN (``INIT_TYPE.upper() == 'SAGAN'``), we follow the
        initialization method in official tensorflow's implementation
        (https://github.com/brain-research/self-attention-gan).

        Besides the reimplementation of the official code's initialization, we
        provide BigGAN's and Pytorch-StudioGAN's style initialization
        (``INIT_TYPE.upper() == BIGGAN`` and ``INIT_TYPE.upper() == STUDIO``).
        Please refer to https://github.com/ajbrock/BigGAN-PyTorch and
        https://github.com/POSTECH-CVLab/PyTorch-StudioGAN.

        Args:
            pretrained (str | dict, optional): Path for the pretrained model or
                dict containing information for pretrained models whose
                necessary key is 'ckpt_path'. Besides, you can also provide
                'prefix' to load the generator part from the whole state dict.
                Defaults to None.
        """
        if isinstance(pretrained, str):
            logger = MMLogger.get_current_instance()
            load_checkpoint(self, pretrained, strict=strict, logger=logger)
        elif isinstance(pretrained, dict):
            ckpt_path = pretrained.get('ckpt_path', None)
            assert ckpt_path is not None
            prefix = pretrained.get('prefix', '')
            map_location = pretrained.get('map_location', 'cpu')
            strict = pretrained.get('strict', True)
            state_dict = _load_checkpoint_with_prefix(prefix, ckpt_path,
                                                      map_location)
            self.load_state_dict(state_dict, strict=strict)
        elif pretrained is None:
            if self.init_type.upper() == 'STUDIO':
                # initialization method from Pytorch-StudioGAN
                #   * weight: orthogonal_init gain=1
                #   * bias  : 0
                for m in self.modules():
                    if isinstance(m, (nn.Conv2d, nn.Linear, nn.Embedding)):
                        nn.init.orthogonal_(m.weight, gain=1)
                        if hasattr(m, 'bias') and m.bias is not None:
                            m.bias.data.fill_(0.)
            elif self.init_type.upper() == 'BIGGAN':
                # initialization method from BigGAN-pytorch
                #   * weight: xavier_init gain=1
                #   * bias  : default
                for m in self.modules():
                    if isinstance(m, (nn.Conv2d, nn.Linear, nn.Embedding)):
                        xavier_uniform_(m.weight, gain=1)
            elif self.init_type.upper() == 'SAGAN':
                # initialization method from official tensorflow code
                #   * weight: xavier_init gain=1
                #   * bias  : 0
                for m in self.modules():
                    if isinstance(m, (nn.Conv2d, nn.Linear, nn.Embedding)):
                        xavier_init(m, gain=1, distribution='uniform')
            elif self.init_type.upper() in ['SNGAN', 'SNGAN-PROJ', 'GAN-PROJ']:
                # initialization method from the official chainer code
                #   * embedding.weight: xavier_init gain=1
                #   * conv.weight     : xavier_init gain=sqrt(2)
                #   * shortcut.weight : xavier_init gain=1
                #   * bias            : 0
                for n, m in self.named_modules():
                    if isinstance(m, nn.Conv2d):
                        if 'shortcut' in n:
                            xavier_init(m, gain=1, distribution='uniform')
                        else:
                            xavier_init(
                                m, gain=np.sqrt(2), distribution='uniform')
                    if isinstance(m, (nn.Linear, nn.Embedding)):
                        xavier_init(m, gain=1, distribution='uniform')
            else:
                raise NotImplementedError('Unknown initialization method: '
                                          f'\'{self.init_type}\'')
        else:
            raise TypeError("'pretrained' must by a str or None. "
                            f'But receive {type(pretrained)}.')



# Copyright (c) OpenMMLab. All rights reserved.
import os
from typing import Dict, List, Optional, Union

import cv2
import mmcv
import numpy as np
import PIL.Image
import requests
import torch
from controlnet_aux import HEDdetector
from mmengine.config import Config
from mmengine.runner import set_random_seed

from mmagic.registry import DIFFUSION_SCHEDULERS, MODELS
from mmagic.utils import ConfigType
from .base_mmagic_inferencer import BaseMMagicInferencer

VIDEO_EXTENSIONS = ('.mp4', '.mov', '.avi')
IMAGE_EXTENSIONS = ('.jpg', '.jpeg', '.png', '.JPG', '.JPEG', '.PNG')


def load_image(image: Union[str, PIL.Image.Image]) -> PIL.Image.Image:
    """
    Args:
    Loads `image` to a PIL Image.
        image (`str` or `PIL.Image.Image`):
            The image to convert to the PIL Image format.
    Returns:
        `PIL.Image.Image`: A PIL Image.
    """
    if isinstance(image, str):
        if image.startswith('http://') or image.startswith('https://'):
            image = PIL.Image.open(requests.get(image, stream=True).raw)
        elif os.path.isfile(image):
            image = PIL.Image.open(image)
        else:
            raise ValueError(
                f'Incorrect path or url, URLs must start with `http://` '
                f'or `https://`, and {image} is not a valid path')
    elif isinstance(image, PIL.Image.Image):
        image = image
    else:
        raise ValueError(
            'Incorrect format used for image. Should be an url linking'
            ' to an image, a local path, or a PIL image.')
    image = PIL.ImageOps.exif_transpose(image)
    image = image.convert('RGB')
    return image


class ControlnetAnimationInferencer(BaseMMagicInferencer):
    """Base inferencer.

    Args:
        config (str or ConfigType): Model config or the path to it.
        ckpt (str, optional): Path to the checkpoint.
        device (str, optional): Device to run inference. If None, the best
            device will be automatically used.
        result_out_dir (str): Output directory of images. Defaults to ''.
    """

    func_kwargs = dict(
        preprocess=[],
        forward=[],
        visualize=['result_out_dir'],
        postprocess=['get_datasample'])
    func_order = dict(preprocess=0, forward=1, visualize=2, postprocess=3)

    extra_parameters = dict()

    def __init__(self,
                 config: Union[ConfigType, str],
                 device: Optional[str] = None,
                 extra_parameters: Optional[Dict] = None,
                 dtype=torch.float32,
                 **kwargs) -> None:
        cfg = Config.fromfile(config)
        self.hed = HEDdetector.from_pretrained(cfg.control_detector)
        self.pipe = MODELS.build(cfg.model).cuda()

        control_scheduler_cfg = dict(
            type=cfg.control_scheduler,
            from_config=self.pipe.scheduler.config,
        )
        control_scheduler = DIFFUSION_SCHEDULERS.build(control_scheduler_cfg)
        self.pipe.test_scheduler = control_scheduler

    @torch.no_grad()
    def __call__(self,
                 prompt=None,
                 video=None,
                 negative_prompt=None,
                 controlnet_conditioning_scale=0.7,
                 image_width=512,
                 image_height=512,
                 save_path=None,
                 strength=0.75,
                 num_inference_steps=20,
                 seed=1,
                 output_fps=None,
                 **kwargs) -> Union[Dict, List[Dict]]:
        """Call the inferencer.

        Args:
            kwargs: Keyword arguments for the inferencer.

        Returns:
            Union[Dict, List[Dict]]: Results of inference pipeline.
        """
        if save_path is None:
            from datetime import datetime
            datestring = datetime.now().strftime('%y%m%d-%H%M%S')
            save_path = '/tmp/' + datestring + '.mp4'

        set_random_seed(seed)

        latent_width = image_width // 8
        latent_height = image_height // 8

        init_noise_shape = (1, 4, latent_height, latent_width)
        init_noise_all_frame = torch.randn(
            init_noise_shape, dtype=self.pipe.controlnet.dtype).cuda()

        init_noise_shape_cat = (1, 4, latent_height, latent_width * 3)
        init_noise_all_frame_cat = torch.randn(
            init_noise_shape_cat, dtype=self.pipe.controlnet.dtype).cuda()

        latent_mask = torch.zeros(
            (1, 4, image_height // 8, image_width // 8 * 3))
        latent_mask[:, :, :,
                    image_width // 8 + 1:image_width // 8 * 2 - 1] = 1.0
        latent_mask = latent_mask.type(self.pipe.controlnet.dtype).cuda()

        # load the images
        input_file_extension = os.path.splitext(video)[1]
        from_video = True
        all_images = []
        if input_file_extension in VIDEO_EXTENSIONS:
            video_reader = mmcv.VideoReader(video)
            input_fps = int(video_reader.fps)
            if output_fps is None:
                output_fps = input_fps
            if output_fps > input_fps:
                output_fps = input_fps
            sample_rate = int(input_fps / output_fps)

            fourcc = cv2.VideoWriter_fourcc(*'mp4v')
            video_writer = cv2.VideoWriter(save_path, fourcc, output_fps,
                                           (image_width, image_height))
            for frame in video_reader:
                all_images.append(np.flip(frame, axis=2))
        else:
            frame_files = os.listdir(video)
            frame_files = [os.path.join(video, f) for f in frame_files]
            frame_files.sort()
            for frame in frame_files:
                frame_extension = os.path.splitext(frame)[1]
                if frame_extension in IMAGE_EXTENSIONS:
                    all_images.append(frame)

            if not os.path.exists(save_path):
                os.makedirs(save_path)

            from_video = False

        # first result
        image = None
        if from_video:
            image = PIL.Image.fromarray(all_images[0])
        else:
            image = load_image(all_images[0])
        image = image.resize((image_width, image_height))
        detect_resolution = min(image_width, image_height)
        hed_image = self.hed(
            image,
            detect_resolution=detect_resolution,
            image_resolution=detect_resolution)
        hed_image = hed_image.resize((image_width, image_height))

        result = self.pipe.infer(
            control=hed_image,
            latent_image=image,
            prompt=prompt,
            negative_prompt=negative_prompt,
            strength=strength,
            controlnet_conditioning_scale=controlnet_conditioning_scale,
            num_inference_steps=num_inference_steps,
            latents=init_noise_all_frame)['samples'][0]

        first_result = result
        first_hed = hed_image
        last_result = result
        last_hed = hed_image

        for ind in range(len(all_images)):
            if from_video:
                if ind % sample_rate > 0:
                    continue
                image = PIL.Image.fromarray(all_images[ind])
            else:
                image = load_image(all_images[ind])
            print('processing frame ind ' + str(ind))

            image = image.resize((image_width, image_height))
            hed_image = self.hed(image, image_resolution=image_width)

            concat_img = PIL.Image.new('RGB', (image_width * 3, image_height))
            concat_img.paste(last_result, (0, 0))
            concat_img.paste(image, (image_width, 0))
            concat_img.paste(first_result, (image_width * 2, 0))

            concat_hed = PIL.Image.new('RGB', (image_width * 3, image_height),
                                       'black')
            concat_hed.paste(last_hed, (0, 0))
            concat_hed.paste(hed_image, (image_width, 0))
            concat_hed.paste(first_hed, (image_width * 2, 0))

            result = self.pipe.infer(
                control=concat_hed,
                latent_image=concat_img,
                prompt=prompt,
                negative_prompt=negative_prompt,
                strength=strength,
                controlnet_conditioning_scale=controlnet_conditioning_scale,
                num_inference_steps=num_inference_steps,
                latents=init_noise_all_frame_cat,
                latent_mask=latent_mask,
            )['samples'][0]
            result = result.crop(
                (image_width, 0, image_width * 2, image_height))

            last_result = result
            last_hed = hed_image

            if from_video:
                video_writer.write(np.flip(np.asarray(result), axis=2))
            else:
                frame_name = frame_files[ind].split('/')[-1]
                save_name = os.path.join(save_path, frame_name)
                result.save(save_name)

        if from_video:
            video_writer.release()

        return save_path



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, Optional, Union

import numpy as np
import torch
import torch.nn as nn
from mmengine import Config, MessageHub
from mmengine.optim import OptimWrapper
from torch import Tensor

from mmagic.registry import MODELS
from mmagic.structures import DataSample
from mmagic.utils.typing import SampleList
from ...utils import get_module_device, get_valid_num_batches
from ..stylegan2 import StyleGAN2
from .stylegan3_utils import (apply_fractional_pseudo_rotation,
                              apply_fractional_rotation,
                              apply_fractional_translation,
                              apply_integer_translation, rotation_matrix)

ModelType = Union[Dict, nn.Module]


@MODELS.register_module()
class StyleGAN3(StyleGAN2):
    """Implementation of `Alias-Free Generative Adversarial Networks`. # noqa.

    Paper link: https://nvlabs-fi-cdn.nvidia.com/stylegan3/stylegan3-paper.pdf # noqa

    Detailed architecture can be found in

    :class:`~mmagic.models.editors.stylegan3.StyleGAN3Generator`
    and
    :class:`~mmagic.models.editors.stylegan2.StyleGAN2Discriminator`
    """

    def __init__(self,
                 generator: ModelType,
                 discriminator: Optional[ModelType] = None,
                 data_preprocessor: Optional[Union[dict, Config]] = None,
                 generator_steps: int = 1,
                 discriminator_steps: int = 1,
                 forward_kwargs: Optional[Dict] = None,
                 ema_config: Optional[Dict] = None,
                 loss_config=dict()):
        super().__init__(generator, discriminator, data_preprocessor,
                         generator_steps, discriminator_steps, ema_config,
                         loss_config)

        self.noise_size = getattr(self.generator, 'noise_size', 512)
        forward_kwargs = dict() if forward_kwargs is None else forward_kwargs
        disc_default_forward_kwargs = dict(update_emas=True, force_fp32=False)
        gen_default_forward_kwargs = dict(force_fp32=False)
        forward_kwargs.setdefault('disc', disc_default_forward_kwargs)
        forward_kwargs.setdefault('gen', gen_default_forward_kwargs)
        self.forward_kwargs = forward_kwargs

    def test_step(self, data: dict) -> SampleList:
        """Gets the generated image of given data. Same as :meth:`val_step`.

        Args:
            data (dict): Data sampled from metric specific
                sampler. More details in `Metrics` and `Evaluator`.

        Returns:
            SampleList: A list of ``DataSample`` contain generated results.
        """
        data = self.data_preprocessor(data)
        inputs_dict, data_samples = data['inputs'], data['data_samples']
        # hard code to compute equivarience
        if 'mode' in inputs_dict and 'eq_cfg' in inputs_dict['mode']:
            batch_size = get_valid_num_batches(inputs_dict, data_samples)
            outputs = self.sample_equivarience_pairs(
                batch_size,
                sample_mode=inputs_dict['mode']['sample_mode'],
                eq_cfg=inputs_dict['mode']['eq_cfg'],
                sample_kwargs=inputs_dict['mode']['sample_kwargs'])
        else:
            outputs = self(inputs_dict, data_samples)
        return outputs

    def val_step(self, data: dict) -> SampleList:
        """Gets the generated image of given data. Same as :meth:`val_step`.

        Args:
            data (dict): Data sampled from metric specific
                sampler. More details in `Metrics` and `Evaluator`.

        Returns:
            SampleList: A list of ``DataSample`` contain generated results.
        """
        data = self.data_preprocessor(data)
        inputs_dict, data_samples = data['inputs'], data['data_samples']
        # hard code to compute equivarience
        if 'mode' in inputs_dict and 'eq_cfg' in inputs_dict['mode']:
            batch_size = get_valid_num_batches(inputs_dict, data_samples)
            outputs = self.sample_equivarience_pairs(
                batch_size,
                sample_mode=inputs_dict['mode']['sample_mode'],
                eq_cfg=inputs_dict['mode']['eq_cfg'],
                sample_kwargs=inputs_dict['mode']['sample_kwargs'])
        else:
            outputs = self(inputs_dict, data_samples)
        return outputs

    def train_discriminator(self, inputs: dict, data_samples: DataSample,
                            optimizer_wrapper: OptimWrapper
                            ) -> Dict[str, Tensor]:
        """Train discriminator.

        Args:
            inputs (dict): Inputs from dataloader.
            data_samples (DataSample): Data samples from dataloader.
            optim_wrapper (OptimWrapper): OptimWrapper instance used to update
                model parameters.
        Returns:
            Dict[str, Tensor]: A ``dict`` of tensor for logging.
        """
        real_imgs = data_samples.gt_img

        num_batches = real_imgs.shape[0]

        noise_batch = self.noise_fn(num_batches=num_batches)
        with torch.no_grad():
            fake_imgs = self.generator(
                noise_batch, return_noise=False, **self.forward_kwargs['disc'])

        disc_pred_fake = self.discriminator(fake_imgs)
        disc_pred_real = self.discriminator(real_imgs)

        parsed_losses, log_vars = self.disc_loss(disc_pred_fake,
                                                 disc_pred_real, real_imgs)
        optimizer_wrapper.update_params(parsed_losses)
        # save ada info
        message_hub = MessageHub.get_current_instance()
        message_hub.update_info('disc_pred_real', disc_pred_real)
        return log_vars

    def train_generator(self, inputs: dict, data_samples: DataSample,
                        optimizer_wrapper: OptimWrapper) -> Dict[str, Tensor]:
        """Train generator.

        Args:
            inputs (dict): Inputs from dataloader.
            data_samples (DataSample): Data samples from dataloader.
                Do not used in generator's training.
            optim_wrapper (OptimWrapper): OptimWrapper instance used to update
                model parameters.

        Returns:
            Dict[str, Tensor]: A ``dict`` of tensor for logging.
        """
        num_batches = len(data_samples)

        noise = self.noise_fn(num_batches=num_batches)
        fake_imgs = self.generator(
            noise, return_noise=False, **self.forward_kwargs['gen'])

        disc_pred_fake = self.discriminator(fake_imgs)
        parsed_loss, log_vars = self.gen_loss(disc_pred_fake, num_batches)

        optimizer_wrapper.update_params(parsed_loss)
        return log_vars

    def sample_equivarience_pairs(self,
                                  batch_size,
                                  sample_mode='ema',
                                  eq_cfg=dict(
                                      compute_eqt_int=False,
                                      compute_eqt_frac=False,
                                      compute_eqr=False,
                                      translate_max=0.125,
                                      rotate_max=1),
                                  sample_kwargs=dict()):
        generator = self.generator if (sample_mode
                                       == 'orig') else self.generator_ema
        if hasattr(generator, 'module'):
            generator = generator.module

        device = get_module_device(generator)
        identity_matrix = torch.eye(3, device=device)

        # Run mapping network.
        z = torch.randn([batch_size, self.noise_size], device=device)
        ws = generator.style_mapping(z=z)
        transform_matrix = getattr(
            getattr(getattr(generator, 'synthesis', None), 'input', None),
            'transform', None)

        # Generate reference image.
        transform_matrix[:] = identity_matrix
        orig = generator.synthesis(ws=ws, **sample_kwargs)

        batch_sample = [DataSample() for _ in range(batch_size)]
        # Integer translation (EQ-T).
        if eq_cfg['compute_eqt_int']:
            t = (torch.rand(2, device=device) * 2 -
                 1) * eq_cfg['translate_max']
            t = (t * generator.out_size).round() / generator.out_size
            transform_matrix[:] = identity_matrix
            transform_matrix[:2, 2] = -t
            img = generator.synthesis(ws=ws, **sample_kwargs)
            ref, mask = apply_integer_translation(orig, t[0], t[1])

            diff = (ref - img).square() * mask
            for idx in range(batch_size):
                data_sample = batch_sample[idx]
                setattr(data_sample, 'eqt_int',
                        DataSample(diff=diff, mask=mask))

        # Fractional translation (EQ-T_frac).
        if eq_cfg['compute_eqt_frac']:
            t = (torch.rand(2, device=device) * 2 -
                 1) * eq_cfg['translate_max']
            transform_matrix[:] = identity_matrix
            transform_matrix[:2, 2] = -t
            img = generator.synthesis(ws=ws, **sample_kwargs)
            ref, mask = apply_fractional_translation(orig, t[0], t[1])

            diff = (ref - img).square() * mask
            for idx in range(batch_size):
                data_sample = batch_sample[idx]
                setattr(data_sample, 'eqt_frac',
                        DataSample(diff=diff, mask=mask))

        # Rotation (EQ-R).
        if eq_cfg['compute_eqr']:
            angle = (torch.rand([], device=device) * 2 - 1) * (
                eq_cfg['rotate_max'] * np.pi)
            transform_matrix[:] = rotation_matrix(-angle)
            img = generator.synthesis(ws=ws, **sample_kwargs)
            ref, ref_mask = apply_fractional_rotation(orig, angle)
            pseudo, pseudo_mask = apply_fractional_pseudo_rotation(img, angle)
            mask = ref_mask * pseudo_mask

            diff = (ref - pseudo).square() * mask
            for idx in range(batch_size):
                data_sample = batch_sample[idx]
                setattr(data_sample, 'eqr', DataSample(diff=diff, mask=mask))

        return batch_sample



# Copyright (c) OpenMMLab. All rights reserved.
import os
import os.path as osp
from collections import defaultdict
from typing import Dict, List, Optional, Sequence, Union

import numpy as np
import torch
from mmengine import print_log
from mmengine.utils import digit_version
from mmengine.utils.dl_utils import TORCH_VERSION
from PIL import Image
from torch.nn import functional as F
from torchvision.utils import make_grid

from mmagic.structures import DataSample
from mmagic.utils import ForwardInputs, try_import
from .base_mmagic_inferencer import BaseMMagicInferencer, InputsType, PredType
from .inference_functions import calculate_grid_size

imageio = try_import('imageio')


class EG3DInferencer(BaseMMagicInferencer):

    func_kwargs = dict(
        preprocess=['inputs'],
        forward=['num_images', 'interpolation'],
        visualize=[
            'result_out_dir', 'vis_mode', 'save_img', 'save_video',
            'img_suffix', 'video_suffix'
        ],
        postprocess=[])

    extra_parameters = dict(num_batches=4, sample_model='ema', add_noise=False)

    def preprocess(self, inputs: InputsType = None) -> ForwardInputs:
        """Process the inputs into a model-feedable format.

        Args:
            inputs (List[Union[str, np.ndarray]]): The conditional inputs for
                the inferencer. Defaults to None.

        Returns:
            ForwardInputs: The preprocessed inputs and data samples.
        """
        if isinstance(inputs, Sequence):
            assert all([type(inputs[0]) == type(lab) for lab in inputs
                        ]), ('All label inputs must have the same type.')
            if isinstance(inputs[0], list):
                for lab in inputs:
                    assert all([isinstance(l_, float) for l_ in lab])
                inputs = np.array(inputs).astype(np.float32)
            elif isinstance(inputs[0], np.ndarray):
                assert all([lab.ndim == 1 for lab in inputs])
                inputs = [input_.astype(np.float32) for input_ in inputs]
            else:
                raise ValueError(
                    'EG3D only support ndarry or list as label input.')

            data_sample_list = []
            for lab in inputs:
                data_sample = DataSample()
                data_sample.set_gt_label(lab)
                data_sample_list.append(data_sample.to(self.device))
            self.extra_parameters['num_batches'] = len(inputs)
        else:
            data_sample_list = None

        num_batches = self.extra_parameters['num_batches']
        sample_model = self.extra_parameters['sample_model']
        add_noise = self.extra_parameters['add_noise']
        inputs = dict(
            num_batches=num_batches,
            sample_model=sample_model,
            add_noise=add_noise)

        if data_sample_list is None:
            data_samples = None
        else:
            data_samples = DataSample.stack(data_sample_list)

        return inputs, data_samples

    def forward(self,
                inputs: ForwardInputs,
                interpolation: Optional[str] = 'both',
                num_images: int = 100) -> Union[dict, List[dict]]:
        """Forward the inputs to the model.

        Args:
            inputs (ForwardInputs): Model inputs. If data sample (the second
                element of `inputs`) is not passed, will generate a sequence
                of images corresponding to passed `interpolation` mode.
            interpolation (str): The interpolation mode. Supported choices
                are 'both', 'conditioning', and 'camera'. Defaults to 'both'.
            num_images (int): The number of frames of interpolation.
                Defaults to 500.

        Returns:
            Union[dict, List[dict]]: Output dict corresponds to the input
                condition or the list of output dict of each frame during the
                interpolation process.
        """
        inputs, data_sample = inputs  # unpack the tuple
        # forward as the passed input
        if data_sample is not None:
            outputs = self.model(inputs, data_sample)
            output_dict = defaultdict(list)
            # return outputs
            for output in outputs:
                fake_img = output.fake_img.data
                depth_img = output.depth
                lr_img = output.lr_img.data
                ray_origins = output.ray_origins
                ray_directions = output.ray_directions
                output_dict['fake_img'].append(fake_img)
                output_dict['depth'].append(depth_img)
                output_dict['lr_img'].append(lr_img)
                output_dict['ray_origins'].append(ray_origins)
                output_dict['ray_directions'].append(ray_directions)

            for k in output_dict.keys():
                output_dict[k] = torch.stack(output_dict[k], dim=0)

            return output_dict

        num_batches = inputs['num_batches']
        output_list = self.model.interpolation(num_images, num_batches,
                                               interpolation)
        return output_list

    def visualize(self,
                  preds: Union[PredType, List[PredType]],
                  vis_mode: str = 'both',
                  save_img: bool = True,
                  save_video: bool = True,
                  img_suffix: str = '.png',
                  video_suffix: str = '.mp4',
                  result_out_dir: str = 'eg3d_output') -> None:
        """Visualize predictions.

        Args:
            preds (Union[PredType, List[PredType]]): Prediction os model.
            vis_mode (str, optional): Which output to visualize. Supported
                choices are 'both', 'depth', and 'img'. Defaults to 'all'.
            save_img (bool, optional): Whether save images. Defaults to True.
            save_video (bool, optional): Whether save videos. Defaults to True.
            img_suffix (str, optional): The suffix of saved images.
                Defaults to '.png'.
            video_suffix (str, optional): The suffix of saved videos.
                Defaults to '.mp4'.
            result_out_dir (str, optional): The save director of image and
                videos. Defaults to 'eg3d_output'.
        """
        if save_video:
            assert imageio is not None, (
                'Please install imageio-ffmpeg by \'pip install '
                'imageio-ffmpeg\' to save video.')

        os.makedirs(result_out_dir, exist_ok=True)
        assert vis_mode.upper() in ['BOTH', 'DEPTH', 'IMG']
        if vis_mode.upper() == 'BOTH':
            vis_mode = ['DEPTH', 'IMG']
        if not isinstance(vis_mode, list):
            vis_mode = [vis_mode]

        if not isinstance(preds, list):
            preds = [preds]
            if save_video:
                save_video = False
                print_log('Only one frame of output is generated and cannot '
                          'save video. Set \'save_video\' to \'False\' '
                          'automatically.')
            if not save_img:
                save_img = True
                print_log('Only one frame of output is generated can only save'
                          'image. Set \'save_img\' to \'True\' automatically.')

        # save video
        batch_size = preds[0]['fake_img'].shape[0]

        img_dict = {}
        for target in vis_mode:
            target = 'fake_img' if target.upper() == 'IMG' else target
            if target.lower() == 'fake_img':
                imgs = self.preprocess_img(preds)
            else:
                imgs = self.preprocess_depth(preds)
            img_dict[target.lower()] = imgs

            nrow = calculate_grid_size(batch_size)

            if save_video:
                video_path = osp.join(
                    result_out_dir,
                    f'{target.lower()}_seed{self.seed}{video_suffix}')
                video_writer = imageio.get_writer(
                    video_path,
                    mode='I',
                    fps=60,
                    codec='libx264',
                    bitrate='10M')

            frame_list = torch.split(imgs, batch_size)
            for idx, frame in enumerate(frame_list):
                # frame: [bz, C, H, W]
                frame_grid = make_grid(
                    frame, nrow=nrow).permute(1, 2, 0)[..., (2, 1, 0)]
                frame_grid = frame_grid.numpy().astype(np.uint8)
                if save_video:
                    video_writer.append_data(frame_grid)

                if save_img:
                    if len(frame_list) != 1:
                        img_name = (f'{target.lower()}_frame{idx}_'
                                    f'seed{self.seed}{img_suffix}')
                    else:
                        img_name = (f'{target.lower()}_seed{self.seed}'
                                    f'{img_suffix}')
                    img_path = osp.join(result_out_dir, img_name)
                    Image.fromarray(frame_grid).save(img_path)

            if save_video:
                video_writer.close()
                print_log(f'Save video to \'{video_path}\'.', 'current')

        if len(vis_mode) > 1:
            fake_img = img_dict['fake_img']
            depth_img = img_dict['depth']
            # [num_frame * bz, 3, H, W * 2]
            imgs = torch.cat([fake_img, depth_img], dim=-1)
            nrow = calculate_grid_size(batch_size, aspect_ratio=2)

            if save_video:
                video_path = osp.join(
                    result_out_dir, f'combine_seed{self.seed}{video_suffix}')
                video_writer = imageio.get_writer(
                    video_path,
                    mode='I',
                    fps=60,
                    codec='libx264',
                    bitrate='10M')

            frame_list = torch.split(imgs, batch_size)
            for idx, frame in enumerate(frame_list):
                frame_grid = make_grid(
                    frame, nrow=nrow).permute(1, 2, 0)[..., (2, 1, 0)]
                frame_grid = frame_grid.numpy().astype(np.uint8)

                if save_video:
                    video_writer.append_data(frame_grid)

                if save_img:
                    if len(frame_list) != 1:
                        img_name = (f'combine_frame{idx}_'
                                    f'seed{self.seed}{img_suffix}')
                    else:
                        img_name = (f'combine_seed{self.seed}' f'{img_suffix}')
                    img_path = osp.join(result_out_dir, img_name)
                    Image.fromarray(frame_grid).save(img_path)

            if save_video:
                video_writer.close()
                print_log(f'Save video to \'{video_path}\'.', 'current')

    def preprocess_img(self, preds: List[dict]) -> torch.Tensor:
        """Preprocess images in the predictions.

        Args:
            preds (List[dict]): List of prediction dict of each frame.

        Returns:
            torch.Tensor: Preprocessed image tensor shape like
                [num_frame * bz, 3, H, W].
        """
        imgs = [p['fake_img'].cpu() for p in preds]
        imgs = torch.cat(imgs, dim=0)  # [num_frame * bz, 3, H, W]
        imgs = ((imgs + 1) / 2 * 255.).clamp(0, 255)
        return imgs

    def preprocess_depth(self, preds: List[dict]) -> torch.Tensor:
        """Preprocess depth in the predictions.

        Args:
            preds (List[dict]): List of prediction dict of each frame.

        Returns:
            torch.Tensor: Preprocessed depth tensor shape like
                [num_frame * bz, 3, H, W].
        """
        depth = [p['depth'].cpu() for p in preds]

        depth = torch.cat(depth, dim=0)
        depth = -depth
        depth = (depth - depth.min()) / (depth.max() - depth.min()) * 255.
        depth = depth.clamp(0, 255).repeat(1, 3, 1, 1)

        img_size = preds[0]['fake_img'].shape[-1]
        if img_size != depth.shape[-1]:
            interpolation_kwargs = dict(
                size=img_size, mode='bilinear', align_corners=False)
            if digit_version(TORCH_VERSION) >= digit_version('1.11.0'):
                interpolation_kwargs['antialias'] = True
            depth = F.interpolate(depth, **interpolation_kwargs)
        return depth

    def postprocess(self,
                    preds: PredType,
                    imgs: Optional[List[np.ndarray]] = None,
                    is_batch: bool = False,
                    get_datasample: bool = False) -> Dict[str, torch.tensor]:
        """Postprocess predictions.

        Args:
            preds (List[Dict]): Predictions of the model.
            imgs (Optional[np.ndarray]): Visualized predictions.
            is_batch (bool): Whether the inputs are in a batch.
                Defaults to False.
            get_datasample (bool): Whether to use Datasample to store
                inference results. If False, dict will be used.

        Returns:
            Dict[str, torch.Tensor]: Inference results as a dict.
        """
        if isinstance(preds[0], dict):
            keys = preds[0].keys()
            outputs = defaultdict(list)
            for pred in preds:
                for k in keys:
                    outputs[k].append(pred[k])
            for k in keys:
                outputs[k] = torch.stack(outputs[k], dim=0)
            return outputs
        # directly return the dict
        return preds



# Copyright (c) OpenMMLab. All rights reserved.
import argparse
import math
import os
import os.path as osp
import sys
from multiprocessing import Pool

import cv2
import lmdb
import mmcv
import mmengine
import numpy as np
from skimage import img_as_float
from skimage.io import imread, imsave

from mmagic.datasets.transforms import MATLABLikeResize, blur_kernels
from mmagic.utils import modify_args


def make_lmdb(mode,
              data_path,
              lmdb_path,
              train_list,
              batch=5000,
              compress_level=1):
    """Create lmdb for the Vimeo90K dataset.

    Contents of lmdb. The file structure is:
    example.lmdb
    ├── data.mdb
    ├── lock.mdb
    ├── meta_info.txt

    The data.mdb and lock.mdb are standard lmdb files and you can refer to
    https://lmdb.readthedocs.io/en/release/ for more details.

    The meta_info.txt is a specified txt file to record the meta information
    of our datasets. It will be automatically created when preparing
    datasets by our provided dataset tools.
    Each line in the txt file records 1)image name (with extension),
    2)image shape, and 3)compression level, separated by a white space.

    For example, the meta information could be:
    `000_00000000.png (720,1280,3) 1`, which means:
    1) image name (with extension): 000_00000000.png;
    2) image shape: (720,1280,3);
    3) compression level: 1

    We use the image name without extension as the lmdb key.

    Args:
        mode (str): Dataset mode. 'gt' or 'lq'.
        data_path (str): Data path for reading images.
        lmdb_path (str): Lmdb save path.
        train_list (str): Train list path for Vimeo90K datasets.
        batch (int): After processing batch images, lmdb commits.
            Default: 5000.
        compress_level (int): Compress level when encoding images. Default: 1.
    """

    print(f'Create lmdb for {data_path}, save to {lmdb_path}...')
    if mode == 'gt':
        h_dst, w_dst = 256, 448
    else:
        h_dst, w_dst = 64, 112

    if osp.exists(lmdb_path):
        print(f'Folder {lmdb_path} already exists. Exit.')
        sys.exit(1)

    print('Reading image path list ...')
    with open(train_list) as f:
        train_list = [line.strip() for line in f]

    all_img_list = []
    keys = []
    for line in train_list:
        folder, sub_folder = line.split('/')
        for j in range(1, 8):
            all_img_list.append(
                osp.join(data_path, folder, sub_folder, f'im{j}.png'))
            keys.append('{}_{}_{}'.format(folder, sub_folder, j))
    all_img_list = sorted(all_img_list)
    keys = sorted(keys)

    if mode == 'gt':  # only read the 4th frame for the gt mode
        print('Only keep the 4th frame for gt mode.')
        all_img_list = [v for v in all_img_list if v.endswith('im4.png')]
        keys = [v for v in keys if v.endswith('_4')]

    # create lmdb environment
    # obtain data size for one image
    img = mmcv.imread(osp.join(data_path, all_img_list[0]), flag='unchanged')
    _, img_byte = cv2.imencode('.png', img,
                               [cv2.IMWRITE_PNG_COMPRESSION, compress_level])
    data_size_per_img = img_byte.nbytes
    print('Data size per image is: ', data_size_per_img)
    data_size = data_size_per_img * len(all_img_list)
    env = lmdb.open(lmdb_path, map_size=data_size * 10)

    # write data to lmdb
    pbar = mmengine.ProgressBar(len(all_img_list))
    txn = env.begin(write=True)
    txt_file = open(osp.join(lmdb_path, 'meta_info.txt'), 'w')
    for idx, (path, key) in enumerate(zip(all_img_list, keys)):
        pbar.update()
        key_byte = key.encode('ascii')
        img = mmcv.imread(osp.join(data_path, path), flag='unchanged')
        h, w, c = img.shape
        _, img_byte = cv2.imencode(
            '.png', img, [cv2.IMWRITE_PNG_COMPRESSION, compress_level])
        assert h == h_dst and w == w_dst and c == 3, (
            f'Wrong shape ({h, w}), should be ({h_dst, w_dst}).')
        txn.put(key_byte, img_byte)
        # write meta information
        txt_file.write(f'{key}.png ({h},{w},{c}) {compress_level}\n')
        if idx % batch == 0:
            txn.commit()
            txn = env.begin(write=True)
    txn.commit()
    env.close()
    txt_file.close()
    print('\nFinish writing lmdb.')


def generate_anno_file(clip_list, file_name='meta_info_Vimeo90K_GT.txt'):
    """Generate anno file for Vimeo90K datasets from the official clip list.

    Args:
        clip_list (str): Clip list path for Vimeo90K datasets.
        file_name (str): Saved file name. Default: 'meta_info_Vimeo90K_GT.txt'.
    """

    print(f'Generate annotation files {file_name}...')
    with open(clip_list) as f:
        lines = [line.rstrip() for line in f]
    txt_file = osp.join(osp.dirname(clip_list), file_name)
    with open(txt_file, 'w') as f:
        for line in lines:
            f.write(f'{line} 7 (256, 448, 3)\n')


def imresize(img_path, output_path, scale=None, output_shape=None):
    """Resize the image using MATLAB-like downsampling.

    Args:
        img_path (str): Input image path.
        output_path (str): Output image path.
        scale (float | None, optional): The scale factor of the resize
            operation. If None, it will be determined by output_shape.
            Default: None.
        output_shape (tuple(int) | None, optional): The size of the output
            image. If None, it will be determined by scale. Note that if
            scale is provided, output_shape will not be used.
            Default: None.
    """

    matlab_resize = MATLABLikeResize(
        keys=['data'], scale=scale, output_shape=output_shape)
    img = imread(img_path)
    img = img_as_float(img)
    data = {'data': img}
    output = matlab_resize(data)['data']
    output = np.clip(output, 0.0, 1.0) * 255
    output = np.around(output).astype(np.uint8)
    imsave(output_path, output)


def mesh_grid(kernel_size):
    """Generate the mesh grid, centering at zero.

    Args:
        kernel_size (int): The size of the kernel.

    Returns:
        xy_grid (np.ndarray): stacked xy coordinates with shape
            (kernel_size, kernel_size, 2).
    """
    range_ = np.arange(-(kernel_size - 1.) / 2., (kernel_size - 1.) / 2. + 1.)
    x_grid, y_grid = np.meshgrid(range_, range_)
    xy_grid = np.hstack((x_grid.reshape((kernel_size * kernel_size, 1)),
                         y_grid.reshape(kernel_size * kernel_size,
                                        1))).reshape(kernel_size, kernel_size,
                                                     2)

    return xy_grid


def bd_downsample(img_path, output_path, sigma=1.6, scale=4):
    """Downsampling using BD degradation(Gaussian blurring and downsampling).

    Args:
        img_path (str): Input image path.
        output_path (str): Output image path.
        sigma (float): The sigma of Gaussian blurring kernel. Default: 1.6.
        scale (int): The scale factor of the downsampling. Default: 4.
    """

    # Gaussian blurring
    kernelsize = math.ceil(sigma * 3) * 2 + 2
    kernel = blur_kernels.bivariate_gaussian(
        kernelsize, sigma, grid=mesh_grid(kernelsize))
    img = cv2.imread(img_path)
    img = img_as_float(img)
    output = cv2.filter2D(
        img,
        -1,
        kernel,
        anchor=((kernelsize - 1) // 2, (kernelsize - 1) // 2),
        borderType=cv2.BORDER_REPLICATE)

    # downsampling
    output = output[int(scale / 2) - 1:-int(scale / 2) + 1:scale,
                    int(scale / 2) - 1:-int(scale / 2) + 1:scale, :]

    output = np.clip(output, 0.0, 1.0) * 255
    output = output.astype(np.float32)
    output = np.floor(output + 0.5)
    cv2.imwrite(output_path, output)


def worker(clip_path, args):
    """Worker for each process.

    Args:
        clip_name (str): Path of the clip.

    Returns:
        process_info (str): Process information displayed in progress bar.
    """

    gt_dir = osp.join(args.data_root, 'GT', clip_path)
    bi_dir = osp.join(args.data_root, 'BIx4', clip_path)
    bd_dir = osp.join(args.data_root, 'BDx4', clip_path)
    mmengine.utils.mkdir_or_exist(bi_dir)
    mmengine.utils.mkdir_or_exist(bd_dir)

    img_list = sorted(os.listdir(gt_dir))
    for img in img_list:
        imresize(osp.join(gt_dir, img), osp.join(bi_dir, img), scale=1 / 4)
        bd_downsample(osp.join(gt_dir, img), osp.join(bd_dir, img))

    process_info = f'Processing {clip_path} ...'
    return process_info


def downsample_images(args):
    """Downsample images."""

    clip_list = []
    gt_dir = osp.join(args.data_root, 'GT')
    sequence_list = sorted(os.listdir(gt_dir))
    for sequence in sequence_list:
        sequence_root = osp.join(gt_dir, sequence)
        clip_list.extend(
            [osp.join(sequence, i) for i in sorted(os.listdir(sequence_root))])

    prog_bar = mmengine.ProgressBar(len(clip_list))
    pool = Pool(args.n_thread)
    for path in clip_list:
        pool.apply_async(
            worker, args=(path, args), callback=lambda arg: prog_bar.update())
    pool.close()
    pool.join()
    print('All processes done.')


def parse_args():
    modify_args()
    parser = argparse.ArgumentParser(
        description='Preprocess Vimeo90K datasets',
        epilog='You can download the Vimeo90K dataset '
        'from：http://toflow.csail.mit.edu/')
    parser.add_argument('--data-root', help='dataset root')
    parser.add_argument(
        '--n-thread',
        nargs='?',
        default=8,
        type=int,
        help='thread number when using multiprocessing')
    parser.add_argument(
        '--train_list',
        default=None,
        help='official training list path for Vimeo90K')
    parser.add_argument('--gt-path', default=None, help='GT path for Vimeo90K')
    parser.add_argument('--lq-path', default=None, help='LQ path for Vimeo90K')
    parser.add_argument(
        '--make-lmdb', action='store_true', help='create lmdb files')

    args = parser.parse_args()
    return args


if __name__ == '__main__':
    args = parse_args()

    # generate BIx4 and BDx4
    downsample_images(args)

    # generate image list anno file
    generate_anno_file(
        osp.join(args.data_root, 'sep_trainlist.txt'),
        'meta_info_Vimeo90K_train_GT.txt')
    generate_anno_file(
        osp.join(args.data_root, 'sep_testlist.txt'),
        'meta_info_Vimeo90K_test_GT.txt')

    # create lmdb files
    if args.make_lmdb:
        if args.gt_path is None or args.lq_path is None:
            raise ValueError('gt_path and lq_path cannot be None when '
                             'when creating lmdb files.')
        # create lmdb for gt
        lmdb_path = osp.join(
            osp.dirname(args.gt_path), 'vimeo90k_train_GT.lmdb')
        make_lmdb(
            mode='gt',
            data_path=args.gt_path,
            lmdb_path=lmdb_path,
            train_list=args.train_list)
        # create lmdb for lq
        lmdb_path = osp.join(
            osp.dirname(args.lq_path), 'vimeo90k_train_LR7frames.lmdb')
        make_lmdb(
            mode='lq',
            data_path=args.lq_path,
            lmdb_path=lmdb_path,
            train_list=args.train_list)



# Copyright (c) OpenMMLab. All rights reserved.
from logging import ERROR
from typing import Optional

import torch.nn as nn
from mmengine import print_log
from mmengine.runner import save_checkpoint


def change_base_model(controlnet: nn.Module,
                      curr_model: nn.Module,
                      base_model: nn.Module,
                      save_path: Optional[str] = None,
                      *args,
                      **kwargs) -> nn.Module:
    """This function is used to change the base model of ControlNet. Refers to
    https://github.com/lllyasviel/ControlNet/blob/main/tool_transfer_control.py
    .

    # noqa.

    Args:
        controlnet (nn.Module): The model for ControlNet to convert.
        curr_model (nn.Module): The model of current Stable Diffusion's Unet.
        base_model (nn.Module): The model of Stable Diffusion's Unet which
            ControlNet initialized with.
        save_path (str, optional): The path to save the converted model.
            Defaults to None.

        *args, **kwargs: Arguments for `save_checkpoint`.
    """
    dtype = next(controlnet.parameters()).dtype
    base_state_dict = base_model.state_dict()
    curr_state_dict = curr_model.state_dict()

    print_log('Start convert ControlNet to new Unet.', 'current')
    for k, v in controlnet.state_dict().items():
        if k in base_state_dict:
            base_v = base_state_dict[k].cpu()
            curr_v = curr_state_dict[k].cpu()
            try:
                offset = v.cpu() - base_v
                new_v = offset + curr_v
                controlnet.state_dict()[k].data.copy_(new_v.to(dtype))
                print_log(f'Convert success: \'{k}\'.', 'current')
            except Exception as exception:
                print_log(
                    f'Error occurs when convert \'{k}\'. '
                    'Please check that the model structure of '
                    '\'ControlNet\', \'BaseModel\' and \'CurrentModel\' '
                    'are consistent.', 'current', ERROR)
                raise exception
    if save_path:
        save_checkpoint(controlnet.state_dict(), save_path, *args, **kwargs)

    return controlnet



# Copyright (c) OpenMMLab. All rights reserved.
import argparse

import torch
from mmengine import Config
from mmengine.registry import init_default_scope

from mmagic.registry import MODELS

try:
    from mmengine.analysis import get_model_complexity_info
except ImportError:
    raise ImportError('Please upgrade mmengine >= 0.6.0')


def parse_args():
    parser = argparse.ArgumentParser(
        description='Get a editor complexity',
        formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('config', help='train config file path')
    parser.add_argument(
        '--shape',
        type=int,
        nargs='+',
        default=[3, 250, 250],
        help='Input shape. Supported tasks:\n'
        'Image Super-Resolution: --shape 3 h w\n'
        'Video Super-Resolution: --shape t 3 h w\n'
        'Video Interpolation: --shape t 3 h w\n'
        'Image Restoration: --shape 3 h w\n'
        'Inpainting: --shape 4 h w\n'
        'Matting: --shape 4 h w\n'
        'Unconditional GANs: --shape noisy_size\n'
        'Image Translation: --shape 3 h w')
    parser.add_argument(
        '--activations',
        action='store_true',
        help='Whether to show the Activations')
    parser.add_argument(
        '--out-table',
        action='store_true',
        help='Whether to show the complexity table')
    parser.add_argument(
        '--out-arch',
        action='store_true',
        help='Whether to show the complexity arch')
    args = parser.parse_args()
    return args


def main():
    """
    Examples:

    Image Super-Resolution:
    `python tools/analysis_tools/get_flops.py configs/srcnn/srcnn_x4k915_1xb16-1000k_div2k.py --shape 3 250 250` # noqa

    Video Super-Resolution:
    `python tools/analysis_tools/get_flops.py configs/edvr/edvrm_8xb4-600k_reds.py --shape 5 3 256 256` # noqa

    Video Interpolation:
    `python tools/analysis_tools/get_flops.py configs/flavr/flavr_in4out1_8xb4_vimeo90k-septuplet.py --shape 4 3 64 64` # noqa

    Image Restoration:
    `python tools/analysis_tools/get_flops.py configs/nafnet/nafnet_c64eb11128mb1db1111_8xb8-lr1e-3-400k_gopro.py --shape 3 128 128` # noqa

    Inpainting:
    `python tools/analysis_tools/get_flops.py configs/aot_gan/aot-gan_smpgan_4xb4_places-512x512.py --shape 4 64 64` # noqa

    Matting:
    `python tools/analysis_tools/get_flops.py configs/dim/dim_stage1-v16_1xb1-1000k_comp1k.py --shape 4 256 256` # noqa

    Unconditional GANs:
    `python tools/analysis_tools/get_flops.py configs/wgan-gp/wgangp_GN_1xb64-160kiters_celeba-cropped-128x128.py --shape 128` # noqa

    Image Translation:
    `python tools/analysis_tools/get_flops.py configs/cyclegan/cyclegan_lsgan-id0-resnet-in_1xb1-250kiters_summer2winter.py --shape 3 250 250`
    """

    args = parse_args()

    input_shape = tuple(args.shape)

    cfg = Config.fromfile(args.config)

    init_default_scope(cfg.get('default_scope', 'mmagic'))

    model = MODELS.build(cfg.model)
    inputs = torch.randn(1, *input_shape)
    if torch.cuda.is_available():
        model.cuda()
        inputs = inputs.cuda()
    model.eval()

    if hasattr(model, 'generator'):
        model = model.generator
    elif hasattr(model, 'backbone'):
        model = model.backbone
    if hasattr(model, 'translation'):
        model.forward = model.translation
    elif hasattr(model, 'infer'):
        model.forward = model.infer

    analysis_results = get_model_complexity_info(model, inputs=inputs)
    flops = analysis_results['flops_str']
    params = analysis_results['params_str']
    activations = analysis_results['activations_str']

    split_line = '=' * 30
    print(f'{split_line}\nInput shape: {input_shape}\n'
          f'Flops: {flops}\nParams: {params}\n{split_line}\n')
    if args.activations:
        print(f'Activations: {activations}\n{split_line}\n')
    if args.out_table:
        print(analysis_results['out_table'], '\n')
    if args.out_arch:
        print(analysis_results['out_arch'], '\n')

    if len(input_shape) == 4:
        print('!!!If your network computes N frames in one forward pass, you '
              'may want to divide the FLOPs by N to get the average FLOPs '
              'for each frame.')
    print('!!!Please be cautious if you use the results in papers. '
          'You may need to check if all ops are supported and verify that the '
          'flops computation is correct.')


if __name__ == '__main__':
    main()



# Copyright (c) OpenMMLab. All rights reserved.
import numpy as np
import torch

# Cached construction of constant tensors. Avoids CPU=>GPU copy when the
# same constant is used multiple times.

_constant_cache = dict()


def constant(value, shape=None, dtype=None, device=None, memory_format=None):
    value = np.asarray(value)
    if shape is not None:
        shape = tuple(shape)
    if dtype is None:
        dtype = torch.get_default_dtype()
    if device is None:
        device = torch.device('cpu')
    if memory_format is None:
        memory_format = torch.contiguous_format

    key = (value.shape, value.dtype, value.tobytes(), shape, dtype, device,
           memory_format)
    tensor = _constant_cache.get(key, None)
    if tensor is None:
        tensor = torch.as_tensor(value.copy(), dtype=dtype, device=device)
        if shape is not None:
            tensor, _ = torch.broadcast_tensors(tensor, torch.empty(shape))
        tensor = tensor.contiguous(memory_format=memory_format)
        _constant_cache[key] = tensor
    return tensor



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, List

import torch
from mmengine.optim import OptimWrapperDict

from mmagic.models.base_models import BaseEditModel
from mmagic.models.utils import set_requires_grad
from mmagic.registry import MODELS


@MODELS.register_module()
class SRGAN(BaseEditModel):
    """SRGAN model for single image super-resolution.

    Ref:
    Photo-Realistic Single Image Super-Resolution Using a Generative
    Adversarial Network.

    Args:
        generator (dict): Config for the generator.
        discriminator (dict): Config for the discriminator. Default: None.
        gan_loss (dict): Config for the gan loss.
            Note that the loss weight in gan loss is only for the generator.
        pixel_loss (dict): Config for the pixel loss. Default: None.
        perceptual_loss (dict): Config for the perceptual loss. Default: None.
        train_cfg (dict): Config for training. Default: None.
        test_cfg (dict): Config for testing. Default: None.
        init_cfg (dict, optional): The weight initialized config for
            :class:`BaseModule`. Default: None.
        data_preprocessor (dict, optional): The pre-process config of
            :class:`BaseDataPreprocessor`. Default: None.
    """

    def __init__(self,
                 generator,
                 discriminator=None,
                 gan_loss=None,
                 pixel_loss=None,
                 perceptual_loss=None,
                 train_cfg=None,
                 test_cfg=None,
                 init_cfg=None,
                 data_preprocessor=None):

        super().__init__(
            generator=generator,
            pixel_loss=pixel_loss,
            train_cfg=train_cfg,
            test_cfg=test_cfg,
            init_cfg=init_cfg,
            data_preprocessor=data_preprocessor)

        # discriminator
        self.discriminator = MODELS.build(
            discriminator) if discriminator else None

        # loss
        self.gan_loss = MODELS.build(gan_loss) if gan_loss else None
        self.perceptual_loss = MODELS.build(
            perceptual_loss) if perceptual_loss else None

        self.disc_steps = 1 if self.train_cfg is None else self.train_cfg.get(
            'disc_steps', 1)
        self.disc_repeat = 1 if self.train_cfg is None else self.train_cfg.get(
            'disc_repeat', 1)
        self.disc_init_steps = (0 if self.train_cfg is None else
                                self.train_cfg.get('disc_init_steps', 0))

        self.register_buffer('step_counter', torch.tensor(0), False)

        if self.discriminator is None or self.gan_loss is None:
            # No GAN model or loss.
            self.disc_repeat = 0

    def forward_train(self, inputs, data_samples=None, **kwargs):
        """Forward training. Losses of training is calculated in train_step.

        Args:
            inputs (torch.Tensor): batch input tensor collated by
                :attr:`data_preprocessor`.
            data_samples (List[BaseDataElement], optional):
                data samples collated by :attr:`data_preprocessor`.

        Returns:
            Tensor: Result of ``forward_tensor`` with ``training=True``.
        """

        return self.forward_tensor(
            inputs, data_samples, training=True, **kwargs)

    def forward_tensor(self, inputs, data_samples=None, training=False):
        """Forward tensor. Returns result of simple forward.

        Args:
            inputs (torch.Tensor): batch input tensor collated by
                :attr:`data_preprocessor`.
            data_samples (List[BaseDataElement], optional):
                data samples collated by :attr:`data_preprocessor`.
            training (bool): Whether is training. Default: False.

        Returns:
            Tensor: result of simple forward.
        """

        feats = self.generator(inputs)

        return feats

    def if_run_g(self):
        """Calculates whether need to run the generator step."""

        return (self.step_counter % self.disc_steps == 0
                and self.step_counter >= self.disc_init_steps)

    def if_run_d(self):
        """Calculates whether need to run the discriminator step."""

        return self.discriminator and self.gan_loss

    def g_step(self, batch_outputs: torch.Tensor, batch_gt_data: torch.Tensor):
        """G step of GAN: Calculate losses of generator.

        Args:
            batch_outputs (Tensor): Batch output of generator.
            batch_gt_data (Tensor): Batch GT data.

        Returns:
            dict: Dict of losses.
        """

        losses = dict()

        # pix loss
        if self.pixel_loss:
            losses['loss_pix'] = self.pixel_loss(batch_outputs, batch_gt_data)

        # perceptual loss
        if self.perceptual_loss:
            loss_percep, loss_style = self.perceptual_loss(
                batch_outputs, batch_gt_data)
            if loss_percep is not None:
                losses['loss_perceptual'] = loss_percep
            if loss_style is not None:
                losses['loss_style'] = loss_style

        # gan loss for generator
        if self.gan_loss and self.discriminator:
            fake_g_pred = self.discriminator(batch_outputs)
            losses['loss_gan'] = self.gan_loss(
                fake_g_pred, target_is_real=True, is_disc=False)

        return losses

    def d_step_real(self, batch_outputs, batch_gt_data: torch.Tensor):
        """Real part of D step.

        Args:
            batch_outputs (Tensor): Batch output of generator.
            batch_gt_data (Tensor): Batch GT data.

        Returns:
            Tensor: Real part of gan_loss for discriminator.
        """

        # real
        real_d_pred = self.discriminator(batch_gt_data)
        loss_d_real = self.gan_loss(
            real_d_pred, target_is_real=True, is_disc=True)

        return loss_d_real

    def d_step_fake(self, batch_outputs: torch.Tensor, batch_gt_data):
        """Fake part of D step.

        Args:
            batch_outputs (Tensor): Batch output of generator.
            batch_gt_data (Tensor): Batch GT data.

        Returns:
            Tensor: Fake part of gan_loss for discriminator.
        """

        # fake
        fake_d_pred = self.discriminator(batch_outputs.detach())
        loss_d_fake = self.gan_loss(
            fake_d_pred, target_is_real=False, is_disc=True)

        return loss_d_fake

    def g_step_with_optim(self, batch_outputs: torch.Tensor,
                          batch_gt_data: torch.Tensor,
                          optim_wrapper: OptimWrapperDict):
        """G step with optim of GAN: Calculate losses of generator and run
        optim.

        Args:
            batch_outputs (Tensor): Batch output of generator.
            batch_gt_data (Tensor): Batch GT data.
            optim_wrapper (OptimWrapperDict): Optim wrapper dict.

        Returns:
            dict: Dict of parsed losses.
        """

        g_optim_wrapper = optim_wrapper['generator']

        with g_optim_wrapper.optim_context(self):
            losses_g = self.g_step(batch_outputs, batch_gt_data)

        parsed_losses_g, log_vars_g = self.parse_losses(losses_g)
        g_optim_wrapper.update_params(parsed_losses_g)

        return log_vars_g

    def d_step_with_optim(self, batch_outputs: torch.Tensor,
                          batch_gt_data: torch.Tensor,
                          optim_wrapper: OptimWrapperDict):
        """D step with optim of GAN: Calculate losses of discriminator and run
        optim.

        Args:
            batch_outputs (Tensor): Batch output of generator.
            batch_gt_data (Tensor): Batch GT data.
            optim_wrapper (OptimWrapperDict): Optim wrapper dict.

        Returns:
            dict: Dict of parsed losses.
        """

        log_vars = dict()
        d_optim_wrapper = optim_wrapper['discriminator']

        with d_optim_wrapper.optim_context(self):
            loss_d_real = self.d_step_real(batch_outputs, batch_gt_data)

        parsed_losses_dr, log_vars_dr = self.parse_losses(
            dict(loss_d_real=loss_d_real))
        log_vars.update(log_vars_dr)
        loss_dr = d_optim_wrapper.scale_loss(parsed_losses_dr)
        d_optim_wrapper.backward(loss_dr)

        with d_optim_wrapper.optim_context(self):
            loss_d_fake = self.d_step_fake(batch_outputs, batch_gt_data)

        parsed_losses_df, log_vars_df = self.parse_losses(
            dict(loss_d_fake=loss_d_fake))
        log_vars.update(log_vars_df)
        loss_df = d_optim_wrapper.scale_loss(parsed_losses_df)
        d_optim_wrapper.backward(loss_df)

        if d_optim_wrapper.should_update():
            d_optim_wrapper.step()
            d_optim_wrapper.zero_grad()

        return log_vars

    def extract_gt_data(self, data_samples):
        """extract gt data from data samples.

        Args:
            data_samples (list): List of DataSample.

        Returns:
            Tensor: Extract gt data.
        """

        batch_gt_data = data_samples.gt_img

        return batch_gt_data

    def train_step(self, data: List[dict],
                   optim_wrapper: OptimWrapperDict) -> Dict[str, torch.Tensor]:
        """Train step of GAN-based method.

        Args:
            data (List[dict]): Data sampled from dataloader.
            optim_wrapper (OptimWrapper): OptimWrapper instance
                used to update model parameters.

        Returns:
            Dict[str, torch.Tensor]: A ``dict`` of tensor for logging.
        """

        g_optim_wrapper = optim_wrapper['generator']

        data = self.data_preprocessor(data, True)
        batch_inputs = data['inputs']
        data_samples = data['data_samples']
        batch_gt_data = self.extract_gt_data(data_samples)

        log_vars = dict()

        with g_optim_wrapper.optim_context(self):
            batch_outputs = self.forward_train(batch_inputs, data_samples)

        if self.if_run_g():
            set_requires_grad(self.discriminator, False)

            log_vars_d = self.g_step_with_optim(
                batch_outputs=batch_outputs,
                batch_gt_data=batch_gt_data,
                optim_wrapper=optim_wrapper)

            log_vars.update(log_vars_d)

        if self.if_run_d():
            set_requires_grad(self.discriminator, True)

            for _ in range(self.disc_repeat):
                # detach before function call to resolve PyTorch2.0 compile bug
                log_vars_d = self.d_step_with_optim(
                    batch_outputs=batch_outputs.detach(),
                    batch_gt_data=batch_gt_data,
                    optim_wrapper=optim_wrapper)

            log_vars.update(log_vars_d)

        if 'loss' in log_vars:
            log_vars.pop('loss')

        self.step_counter += 1

        return log_vars



# Copyright (c) OpenMMLab. All rights reserved.
from typing import List

from mmagic.models.base_models import OneStageInpaintor
from mmagic.registry import MODELS


@MODELS.register_module()
class PConvInpaintor(OneStageInpaintor):
    """Inpaintor for Partial Convolution method.

    This inpaintor is implemented according to the paper: Image inpainting for
    irregular holes using partial convolutions
    """

    def forward_tensor(self, inputs, data_samples):
        """Forward function in tensor mode.

        Args:
            inputs (torch.Tensor): Input tensor.
            data_sample (dict): Dict contains data sample.

        Returns:
            dict: Dict contains output results.
        """

        masked_img = inputs  # N,3,H,W
        masks = data_samples.mask
        masks = 1. - masks
        masks = masks.repeat(1, 3, 1, 1)
        fake_reses, _ = self.generator(masked_img, masks)
        fake_imgs = fake_reses * (1. - masks) + masked_img * masks
        return fake_reses, fake_imgs

    def train_step(self, data: List[dict], optim_wrapper):
        """Train step function.

        In this function, the inpaintor will finish the train step following
        the pipeline:

            1. get fake res/image
            2. optimize discriminator (if have)
            3. optimize generator

        If `self.train_cfg.disc_step > 1`, the train step will contain multiple
        iterations for optimizing discriminator with different input data and
        only one iteration for optimizing generator after `disc_step`
        iterations for discriminator.

        Args:
            data (List[dict]): Batch of data as input.
            optim_wrapper (dict[torch.optim.Optimizer]): Dict with optimizers
                for generator and discriminator (if have).

        Returns:
            dict: Dict with loss, information for logger, the number of \
                samples and results for visualization.
        """
        data = self.data_preprocessor(data, True)
        batch_inputs, data_samples = data['inputs'], data['data_samples']
        log_vars = {}

        masked_img = batch_inputs  # float
        gt_img = data_samples.gt_img
        mask = data_samples.mask
        mask = mask.float()

        mask_input = mask.expand_as(gt_img)
        mask_input = 1. - mask_input

        fake_res, final_mask = self.generator(masked_img, mask_input)
        fake_img = gt_img * (1. - mask) + fake_res * mask

        results, g_losses = self.generator_loss(fake_res, fake_img, gt_img,
                                                mask, masked_img)
        loss_g_, log_vars_g = self.parse_losses(g_losses)
        log_vars.update(log_vars_g)
        optim_wrapper.zero_grad()
        optim_wrapper.backward(loss_g_)
        optim_wrapper.step()

        return log_vars



# MMagic's implementation of Stable Diffusion
# unet = dict(
#     type='DenoisingUnet',
#     image_size=512,
#     base_channels=320,
#     channels_cfg=[1, 2, 4, 4],
#     unet_type='stable',
#     act_cfg=dict(type='silu'),
#     cross_attention_dim=768,
#     num_heads=8,
#     in_channels=4,
#     layers_per_block=2,
#     down_block_types=[
#         'CrossAttnDownBlock2D',
#         'CrossAttnDownBlock2D',
#         'CrossAttnDownBlock2D',
#         'DownBlock2D',
#     ],
#     up_block_types=[
#         'UpBlock2D',
#         'CrossAttnUpBlock2D',
#         'CrossAttnUpBlock2D',
#         'CrossAttnUpBlock2D',
#     ],
#     output_cfg=dict(var='fixed'))

# vae = dict(
#     type='EditAutoencoderKL',
#     act_fn='silu',
#     block_out_channels=[128, 256, 512, 512],
#     down_block_types=[
#         'DownEncoderBlock2D',
#         'DownEncoderBlock2D',
#         'DownEncoderBlock2D',
#         'DownEncoderBlock2D',
#     ],
#     in_channels=3,
#     latent_channels=4,
#     layers_per_block=2,
#     norm_num_groups=32,
#     out_channels=3,
#     sample_size=512,
#     up_block_types=[
#         'UpDecoderBlock2D',
#         'UpDecoderBlock2D',
#         'UpDecoderBlock2D',
#         'UpDecoderBlock2D',
#     ])

# Use DiffuserWrapper!
stable_diffusion_v15_url = 'runwayml/stable-diffusion-v1-5'
unet = dict(
    type='UNet2DConditionModel',
    subfolder='unet',
    from_pretrained=stable_diffusion_v15_url)
vae = dict(
    type='AutoencoderKL',
    from_pretrained=stable_diffusion_v15_url,
    subfolder='vae')

diffusion_scheduler = dict(
    type='EditDDIMScheduler',
    variance_type='learned_range',
    beta_end=0.012,
    beta_schedule='scaled_linear',
    beta_start=0.00085,
    num_train_timesteps=1000,
    set_alpha_to_one=False,
    clip_sample=False)

model = dict(
    type='StableDiffusion',
    unet=unet,
    vae=vae,
    enable_xformers=False,
    text_encoder=dict(
        type='ClipWrapper',
        clip_type='huggingface',
        pretrained_model_name_or_path=stable_diffusion_v15_url,
        subfolder='text_encoder'),
    tokenizer=stable_diffusion_v15_url,
    scheduler=diffusion_scheduler,
    test_scheduler=diffusion_scheduler)



# Copyright (c) 2021, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.
"""Custom replacement for `torch.nn.functional.grid_sample` that supports
arbitrarily high order gradients between the input and output.

Only works on 2D images and assumes `mode='bilinear'`, `padding_mode='zeros'`,
`align_corners=False`.
"""

import warnings

import torch

# pylint: disable=redefined-builtin
# pylint: disable=arguments-differ
# pylint: disable=protected-access

# ----------------------------------------------------------------------------

enabled = True  # Enable the custom op by setting this to true.

# ----------------------------------------------------------------------------


def grid_sample(input, grid):
    if _should_use_custom_op():
        return _GridSample2dForward.apply(input, grid)
    return torch.nn.functional.grid_sample(
        input=input,
        grid=grid,
        mode='bilinear',
        padding_mode='zeros',
        align_corners=False)


# ----------------------------------------------------------------------------


def _should_use_custom_op():
    if not enabled:
        return False
    if any(
            torch.__version__.startswith(x)
            for x in ['1.5.', '1.6.', '1.7.', '1.8.', '1.9.', '1.10.']):
        return True
    warnings.warn(
        f'grid_sample_gradfix not supported on PyTorch {torch.__version__}.'
        ' Falling back to torch.nn.functional.grid_sample().')
    return False


# ----------------------------------------------------------------------------


class _GridSample2dForward(torch.autograd.Function):

    @staticmethod
    def forward(ctx, input, grid):
        assert input.ndim == 4
        assert grid.ndim == 4
        output = torch.nn.functional.grid_sample(
            input=input,
            grid=grid,
            mode='bilinear',
            padding_mode='zeros',
            align_corners=False)
        ctx.save_for_backward(input, grid)
        return output

    @staticmethod
    def backward(ctx, grad_output):
        input, grid = ctx.saved_tensors
        grad_input, grad_grid = _GridSample2dBackward.apply(
            grad_output, input, grid)
        return grad_input, grad_grid


# ----------------------------------------------------------------------------


class _GridSample2dBackward(torch.autograd.Function):

    @staticmethod
    def forward(ctx, grad_output, input, grid):
        op = torch._C._jit_get_operation('aten::grid_sampler_2d_backward')
        grad_input, grad_grid = op(grad_output, input, grid, 0, 0, False)
        ctx.save_for_backward(grid)
        return grad_input, grad_grid

    @staticmethod
    def backward(ctx, grad2_grad_input, grad2_grad_grid):
        _ = grad2_grad_grid  # unused
        grid, = ctx.saved_tensors
        grad2_grad_output = None
        grad2_input = None
        grad2_grid = None

        if ctx.needs_input_grad[0]:
            grad2_grad_output = _GridSample2dForward.apply(
                grad2_grad_input, grid)

        assert not ctx.needs_input_grad[2]
        return grad2_grad_output, grad2_input, grad2_grid



# Copyright (c) OpenMMLab. All rights reserved.
import json
import os
from typing import Callable, List, Union

from mmengine.dataset import BaseDataset

from mmagic.registry import DATASETS


@DATASETS.register_module()
class ControlNetDataset(BaseDataset):
    """Demo dataset to test ControlNet. Modified from https://github.com/lllyas
    viel/ControlNet/blob/16ea3b5379c1e78a4bc8e3fc9cae8d65c42511b1/tutorial_data
    set.py  # noqa.

    You can download the demo data from https://huggingface.co/lllyasviel/ControlNet/blob/main/training/fill50k.zip  # noqa
    and then unzip the file to the ``data`` folder.

    Args:
        ann_file (str): Path to the annotation file. Defaults
            to 'prompt.json' as ControlNet's default.
        data_root (str): Path to the data root. Defaults to './data/fill50k'.
        pipeline (list[dict | callable]): A sequence of data transforms.
    """

    def __init__(self,
                 ann_file: str = 'prompt.json',
                 data_root: str = './data/fill50k',
                 control_key='source',
                 image_key='target',
                 pipeline: List[Union[dict, Callable]] = []):
        self.control_key = control_key
        self.image_key = image_key
        super().__init__(
            ann_file=ann_file, data_root=data_root, pipeline=pipeline)

    def load_data_list(self) -> List[dict]:
        """Load annotations from an annotation file named as ``self.ann_file``

        Returns:
            list[dict]: A list of annotation.
        """
        data_list = []
        with open(self.ann_file, 'rt') as file:
            anno_list = file.readlines()

        for anno in anno_list:
            anno = json.loads(anno)
            # source = anno['source']
            # target = anno['target']
            source = anno[self.control_key]
            target = anno[self.image_key]
            prompt = anno['prompt']

            source = os.path.join(self.data_root, source)
            target = os.path.join(self.data_root, target)

            data_list.append({
                'source_path': source,
                'target_path': target,
                'prompt': prompt
            })

        return data_list



# Copyright (c) OpenMMLab. All rights reserved.
from copy import deepcopy
from typing import Dict, List, Optional, Union

import torch
import torch.nn as nn
from mmengine import Config
from mmengine.utils import ProgressBar
from torch import Tensor

from mmagic.registry import MODELS
from mmagic.structures import DataSample
from mmagic.utils.typing import ForwardInputs, SampleList
from ...base_models import BaseConditionalGAN
from ...utils import get_valid_num_batches

ModelType = Union[Dict, nn.Module]


@MODELS.register_module()
class EG3D(BaseConditionalGAN):
    """Implementation of `Efficient Geometry-aware 3D Generative Adversarial
    Networks`

    <https://openaccess.thecvf.com/content/CVPR2022/papers/Chan_Efficient_Geometry-Aware_3D_Generative_Adversarial_Networks_CVPR_2022_paper.pdf>_ (EG3D).  # noqa

    Detailed architecture can be found in
    :class:`~mmagic.models.editors.eg3d.eg3d_generator.TriplaneGenerator`
    and
    :class:`~mmagic.models.editors.eg3d.dual_discriminator.DualDiscriminator`

    Args:
        generator (ModelType): The config or model of the generator.
        discriminator (Optional[ModelType]): The config or model of the
            discriminator. Defaults to None.
        camera (Optional[ModelType]): The pre-defined camera to sample random
            camera position. If you want to generate images or videos via
            high-level API, you must set this argument. Defaults to None.
        data_preprocessor (Optional[Union[dict, Config]]): The pre-process
            config or :class:`~mmagic.models.DataPreprocessor`.
        generator_steps (int): Number of times the generator was completely
            updated before the discriminator is updated. Defaults to 1.
        discriminator_steps (int): Number of times the discriminator was
            completely updated before the generator is updated. Defaults to 1.
        noise_size (Optional[int]): Size of the input noise vector.
            Default to 128.
        num_classes (Optional[int]): The number classes you would like to
            generate. Defaults to None.
        ema_config (Optional[Dict]): The config for generator's exponential
            moving average setting. Defaults to None.
        loss_config (Optional[Dict]): The config for training losses.
            Defaults to None.
    """

    def __init__(self,
                 generator: ModelType,
                 discriminator: Optional[ModelType] = None,
                 camera: Optional[ModelType] = None,
                 data_preprocessor: Optional[Union[dict, Config]] = None,
                 generator_steps: int = 1,
                 discriminator_steps: int = 1,
                 noise_size: Optional[int] = None,
                 ema_config: Optional[Dict] = None,
                 loss_config: Optional[Dict] = None):

        super().__init__(generator, discriminator, data_preprocessor,
                         generator_steps, discriminator_steps, noise_size,
                         None, ema_config, loss_config)
        if isinstance(camera, dict):
            self.camera = MODELS.build(camera)
        elif isinstance(camera, nn.Module):
            self.camera = camera
        else:
            self.camera = None

    def label_fn(self,
                 label: Optional[Tensor] = None,
                 num_batches: int = 1) -> Tensor:
        """Label sampling function for EG3D model.

        Args:
            label (Optional[Tensor]): Conditional for EG3D model. If not
                passed, :attr:`self.camera` will be used to sample random
                camera-to-world and intrinsics matrix. Defaults to None.

        Returns:
            torch.Tensor: Conditional input for EG3D model.
        """
        if label is not None:
            return label

        # sample random conditional from camera
        assert self.camera is not None, (
            '\'camera\' is not defined for \'EG3D\'.')
        camera2world = self.camera.sample_camera2world(batch_size=num_batches)
        intrinsics = self.camera.sample_intrinsic(batch_size=num_batches)
        cond = torch.cat(
            [camera2world.reshape(-1, 16),
             intrinsics.reshape(-1, 9)], dim=1).to(self.device)
        return cond

    def data_sample_to_label(self, data_sample: SampleList
                             ) -> Optional[torch.Tensor]:
        """Get labels from input `data_sample` and pack to `torch.Tensor`. If
        no label is found in the passed `data_sample`, `None` would be
        returned.

        Args:
            data_sample (List[DataSample]): Input data samples.

        Returns:
            Optional[torch.Tensor]: Packed label tensor.
        """

        if not data_sample or 'gt_label' not in data_sample.keys():
            return None
        return data_sample.gt_label.label

    def pack_to_data_sample(
            self,
            output: Dict[str, Tensor],
            # index: int,
            data_sample: Optional[DataSample] = None) -> DataSample:
        """Pack output to data sample. If :attr:`data_sample` is not passed, a
        new DataSample will be instantiated. Otherwise, outputs will be added
        to the passed datasample.

        Args:
            output (Dict[Tensor]): Output of the model.
            index (int): The index to save.
            data_sample (DataSample, optional): Data sample to save
                outputs. Defaults to None.

        Returns:
            DataSample: Data sample with packed outputs.
        """

        assert isinstance(output,
                          dict), ('Output of EG3D generator should be a dict.')

        data_sample = DataSample() if data_sample is None else data_sample
        for k, v in output.items():
            assert isinstance(v, torch.Tensor), (
                f'Output must be tensor. But \'{k}\' is type of '
                f'\'{type(v)}\'.')
            # NOTE: hard code here, we assume all tensor are [bz, ...]
            data_sample.set_tensor_data({k: v})

        return data_sample

    def forward(self,
                inputs: ForwardInputs,
                data_samples: Optional[list] = None,
                mode: Optional[str] = None) -> List[DataSample]:
        """Sample images with the given inputs. If forward mode is 'ema' or
        'orig', the image generated by corresponding generator will be
        returned. If forward mode is 'ema/orig', images generated by original
        generator and EMA generator will both be returned in a dict.

        Args:
            inputs (ForwardInputs): Dict containing the necessary
                information (e.g. noise, num_batches, mode) to generate image.
            data_samples (Optional[list]): Data samples collated by
                :attr:`data_preprocessor`. Defaults to None.
            mode (Optional[str]): `mode` is not used in
                :class:`BaseConditionalGAN`. Defaults to None.

        Returns:
            List[DataSample]: Generated images or image dict.
        """
        if isinstance(inputs, Tensor):
            noise = inputs
            sample_kwargs = {}
        else:
            noise = inputs.get('noise', None)
            num_batches = get_valid_num_batches(inputs, data_samples)
            noise = self.noise_fn(noise, num_batches=num_batches)
            sample_kwargs = inputs.get('sample_kwargs', dict())
        num_batches = noise.shape[0]

        labels = self.data_sample_to_label(data_samples)
        if labels is None:
            num_batches = get_valid_num_batches(inputs, data_samples)
            labels = self.label_fn(num_batches=num_batches)

        sample_model = self._get_valid_model(inputs)
        batch_sample_list = []
        if sample_model in ['ema', 'orig']:
            if sample_model == 'ema':
                generator = self.generator_ema
            else:
                generator = self.generator
            outputs = generator(noise, label=labels)
            outputs['fake_img'] = self.data_preprocessor.destruct(
                outputs['fake_img'], data_samples)

            gen_sample = DataSample()
            if data_samples:
                gen_sample.update(data_samples)
            if isinstance(inputs, dict) and 'img' in inputs:
                gen_sample.gt_img = inputs['img']
            gen_sample = self.pack_to_data_sample(outputs, gen_sample)
            gen_sample.noise = noise
            gen_sample.sample_kwargs = deepcopy(sample_kwargs)
            gen_sample.sample_model = sample_model
            batch_sample_list = gen_sample.split(allow_nonseq_value=True)

        else:
            outputs_orig = self.generator(noise, label=labels)
            outputs_ema = self.generator_ema(noise, label=labels)
            outputs_orig['fake_img'] = self.data_preprocessor.destruct(
                outputs_orig['fake_img'], data_samples)
            outputs_ema['fake_img'] = self.data_preprocessor.destruct(
                outputs_ema['fake_img'], data_samples)

            gen_sample = DataSample()
            if data_samples:
                gen_sample.update(data_samples)
            if isinstance(inputs, dict) and 'img' in inputs:
                gen_sample.gt_img = inputs['img']
            gen_sample.ema = self.pack_to_data_sample(outputs_ema)
            gen_sample.orig = self.pack_to_data_sample(outputs_orig)
            gen_sample.noise = noise
            gen_sample.sample_kwargs = deepcopy(sample_kwargs)
            gen_sample.sample_model = sample_model
            batch_sample_list = gen_sample.split(allow_nonseq_value=True)

        return batch_sample_list

    @torch.no_grad()
    def interpolation(self,
                      num_images: int,
                      num_batches: int = 4,
                      mode: str = 'both',
                      sample_model: str = 'orig',
                      show_pbar: bool = True) -> List[dict]:
        """Interpolation input and return a list of output results. We support
        three kinds of interpolation mode:

        * 'camera': First generate style code with random noise and forward
            camera. Then synthesis images with interpolated camera position
            and fixed style code.

        * 'conditioning': First generate style code with fixed noise and
            interpolated camera. Then synthesis images with style codes and
            forward camera.

        * 'both': Generate images with interpolated camera position.

        Args:
            num_images (int): The number of images want to generate.
            num_batches (int, optional): The number of batches to generate at
                one time. Defaults to 4.
            mode (str, optional): The interpolation mode. Supported choices
                are 'both', 'camera', and 'conditioning'. Defaults to 'both'.
            sample_model (str, optional): The model used to generate images,
                support 'orig' and 'ema'. Defaults to 'orig'.
            show_pbar (bool, optional): Whether display a progress bar during
                interpolation. Defaults to True.

        Returns:
            List[dict]: The list of output dict of each frame.
        """
        assert hasattr(self, 'camera'), ('Camera must be defined.')
        assert mode.upper() in ['BOTH', 'CONDITIONING', 'CAMERA']
        assert sample_model in ['ema', 'orig']
        if sample_model == 'orig':
            gen = self.generator
        else:
            assert self.with_ema_gen, (
                '\'sample_model\' is EMA, but ema model not found.')
            # generator_ema is wrapped by AverageModel
            gen = self.generator_ema.module

        cam2world_forward = self.camera.sample_camera2world(
            h_std=0, v_std=0, batch_size=num_batches, device=self.device)
        intrinsics = self.camera.sample_intrinsic(
            batch_size=num_batches, device=self.device)
        cond_forward = torch.cat([
            cam2world_forward.view(num_batches, -1),
            intrinsics.view(num_batches, -1)
        ],
                                 dim=1)
        # 1. generate cond list
        if mode.upper() in ['CAMERA', 'BOTH']:
            cam2world_list = self.camera.interpolation_cam2world(
                num_images, batch_size=num_batches, device=self.device)
            cond_list = []
            for cam2world in cam2world_list:
                cond = torch.cat([
                    cam2world.view(num_batches, -1),
                    intrinsics.view(num_batches, -1)
                ],
                                 dim=1)
                cond_list.append(cond)
        else:
            cond_list = [cond_forward for _ in range(num_images)]

        # 2. generate pre-defined style list
        if mode.upper() == 'CAMERA':
            # same noise + forward cond
            style = gen.backbone.mapping(
                self.noise_fn(num_batches=num_batches), cond_forward)
            style_list = [style for _ in range(num_images)]
        else:
            # same noise + different cond
            noise = self.noise_fn(num_batches=num_batches)
            style_list = [
                gen.backbone.mapping(noise, cond) for cond in cond_list
            ]

        # 3. interpolation
        if show_pbar:
            pbar = ProgressBar(num_images)
        output_list = []
        for style, cond in zip(style_list, cond_list):
            # generate image with const noise
            output = gen(
                style,
                cond,
                input_is_latent=True,
                add_noise=True,
                randomize_noise=False)  # use fixed noise
            output_list.append({k: v.cpu() for k, v in output.items()})

            if show_pbar:
                pbar.update(1)
        if show_pbar:
            print('\n')

        return output_list



# config for model
stable_diffusion_v15_url = 'Linaqruf/anything-v3.0'
controlnet_hed_url = 'lllyasviel/sd-controlnet-hed'
control_detector = 'lllyasviel/ControlNet'
control_scheduler = 'UniPCMultistepScheduler'

model = dict(
    type='ControlStableDiffusionImg2Img',
    vae=dict(
        type='AutoencoderKL',
        from_pretrained=stable_diffusion_v15_url,
        subfolder='vae'),
    unet=dict(
        type='UNet2DConditionModel',
        subfolder='unet',
        from_pretrained=stable_diffusion_v15_url),
    text_encoder=dict(
        type='ClipWrapper',
        clip_type='huggingface',
        pretrained_model_name_or_path=stable_diffusion_v15_url,
        subfolder='text_encoder'),
    tokenizer=stable_diffusion_v15_url,
    controlnet=dict(
        type='ControlNetModel', from_pretrained=controlnet_hed_url),
    scheduler=dict(
        type='DDPMScheduler',
        from_pretrained=stable_diffusion_v15_url,
        subfolder='scheduler'),
    test_scheduler=dict(
        type='DDIMScheduler',
        from_pretrained=stable_diffusion_v15_url,
        subfolder='scheduler'),
    data_preprocessor=dict(type='DataPreprocessor'),
    init_cfg=dict(type='init_from_unet'),
    enable_xformers=False,
)



_base_ = '../default_runtime.py'
work_dir = './work_dirs/liif'
save_dir = './work_dirs'

scale_min, scale_max = 1, 4
scale_test = 4

train_pipeline = [
    dict(
        type='LoadImageFromFile',
        key='gt',
        color_type='color',
        channel_order='rgb',
        imdecode_backend='cv2'),
    dict(
        type='RandomDownSampling',
        scale_min=scale_min,
        scale_max=scale_max,
        patch_size=48),
    dict(
        type='Flip',
        keys=['img', 'gt'],
        flip_ratio=0.5,
        direction='horizontal'),
    dict(
        type='Flip', keys=['img', 'gt'], flip_ratio=0.5, direction='vertical'),
    dict(type='RandomTransposeHW', keys=['img', 'gt'], transpose_ratio=0.5),
    dict(type='GenerateCoordinateAndCell', sample_quantity=2304),
    dict(type='PackInputs')
]
val_pipeline = [
    dict(
        type='LoadImageFromFile',
        key='gt',
        color_type='color',
        channel_order='rgb',
        imdecode_backend='cv2'),
    dict(type='RandomDownSampling', scale_min=scale_max, scale_max=scale_max),
    dict(type='GenerateCoordinateAndCell', reshape_gt=False),
    dict(type='PackInputs')
]
# test_pipeline = [
#     dict(
#         type='LoadImageFromFile',
#         key='gt',
#         color_type='color',
#         channel_order='rgb',
#         imdecode_backend='cv2'),
#     dict(
#         type='LoadImageFromFile',
#         key='img',
#         color_type='color',
#         channel_order='rgb',
#         imdecode_backend='cv2'),
#     dict(type='GenerateCoordinateAndCell', scale=scale_test,
#          reshape_gt=False),
#     dict(type='PackInputs')
# ]

# dataset settings
dataset_type = 'BasicImageDataset'
data_root = 'data'

train_dataloader = dict(
    num_workers=8,
    batch_size=16,
    persistent_workers=False,
    sampler=dict(type='InfiniteSampler', shuffle=True),
    dataset=dict(
        type=dataset_type,
        ann_file='meta_info_DIV2K800sub_GT.txt',
        metainfo=dict(dataset_type='div2k', task_name='sisr'),
        data_root=data_root + '/DIV2K',
        data_prefix=dict(gt='DIV2K_train_HR_sub'),
        pipeline=train_pipeline))

val_dataloader = dict(
    num_workers=4,
    persistent_workers=False,
    drop_last=False,
    sampler=dict(type='DefaultSampler', shuffle=False),
    dataset=dict(
        type=dataset_type,
        metainfo=dict(dataset_type='set5', task_name='sisr'),
        data_root=data_root + '/Set5',
        data_prefix=dict(img='LRbicx4', gt='GTmod12'),
        pipeline=val_pipeline))

val_evaluator = dict(
    type='Evaluator',
    metrics=[
        dict(type='MAE'),
        dict(type='PSNR', crop_border=scale_max),
        dict(type='SSIM', crop_border=scale_max),
    ])

train_cfg = dict(
    type='IterBasedTrainLoop', max_iters=1_000_000, val_interval=3000)
val_cfg = dict(type='MultiValLoop')

# optimizer
optim_wrapper = dict(
    constructor='DefaultOptimWrapperConstructor',
    type='OptimWrapper',
    optimizer=dict(type='Adam', lr=1e-4))

# learning policy
param_scheduler = dict(
    type='MultiStepLR',
    by_epoch=False,
    milestones=[200_000, 400_000, 600_000, 800_000],
    gamma=0.5)

default_hooks = dict(
    checkpoint=dict(
        type='CheckpointHook',
        interval=3000,
        save_optimizer=True,
        by_epoch=False,
        out_dir=save_dir,
    ),
    timer=dict(type='IterTimerHook'),
    logger=dict(type='LoggerHook', interval=100),
    param_scheduler=dict(type='ParamSchedulerHook'),
    sampler_seed=dict(type='DistSamplerSeedHook'),
)



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, List

import torch
from mmengine.optim import OptimWrapperDict

from mmagic.models.utils import set_requires_grad
from mmagic.registry import MODELS
from ..srgan import SRGAN


@MODELS.register_module()
class DIC(SRGAN):
    """DIC model for Face Super-Resolution.

    Paper: Deep Face Super-Resolution with Iterative Collaboration between
        Attentive Recovery and Landmark Estimation.

    Args:
        generator (dict): Config for the generator.
        pixel_loss (dict): Config for the pixel loss.
        align_loss (dict): Config for the align loss.
        discriminator (dict): Config for the discriminator. Default: None.
        gan_loss (dict): Config for the gan loss. Default: None.
        feature_loss (dict): Config for the feature loss. Default: None.
        train_cfg (dict): Config for train. Default: None.
        test_cfg (dict): Config for testing. Default: None.
        init_cfg (dict, optional): The weight initialized config for
            :class:`BaseModule`. Default: None.
        data_preprocessor (dict, optional): The pre-process config of
            :class:`BaseDataPreprocessor`. Default: None.
    """

    def __init__(self,
                 generator,
                 pixel_loss,
                 align_loss,
                 discriminator=None,
                 gan_loss=None,
                 feature_loss=None,
                 train_cfg=None,
                 test_cfg=None,
                 init_cfg=None,
                 data_preprocessor=None):

        super().__init__(
            generator=generator,
            discriminator=discriminator,
            gan_loss=gan_loss,
            pixel_loss=pixel_loss,
            train_cfg=train_cfg,
            test_cfg=test_cfg,
            init_cfg=init_cfg,
            data_preprocessor=data_preprocessor)

        self.align_loss = MODELS.build(align_loss)
        self.feature_loss = MODELS.build(
            feature_loss) if feature_loss else None

        self.pixel_init = train_cfg.get('pixel_init', 0) if train_cfg else 0

    def forward_tensor(self, inputs, data_samples=None, training=False):
        """Forward tensor. Returns result of simple forward.

        Args:
            inputs (torch.Tensor): batch input tensor collated by
                :attr:`data_preprocessor`.
            data_samples (List[BaseDataElement], optional):
                data samples collated by :attr:`data_preprocessor`.
            training (bool): Whether is training. Default: False.

        Returns:
            (Tensor | Tuple[List[Tensor]]): results of forward inference and
                forward train.
        """

        sr_list, heatmap_list = self.generator(inputs)

        if training:
            return sr_list, heatmap_list
        else:
            return sr_list[-1]

    def if_run_g(self):
        """Calculates whether need to run the generator step."""

        return True

    def if_run_d(self):
        """Calculates whether need to run the discriminator step."""

        return self.step_counter >= self.pixel_init and super().if_run_d()

    def g_step(self, batch_outputs, batch_gt_data):
        """G step of GAN: Calculate losses of generator.

        Args:
            batch_outputs (Tensor): Batch output of generator.
            batch_gt_data (Tensor): Batch GT data.

        Returns:
            dict: Dict of losses.
        """

        sr_list, heatmap_list = batch_outputs
        gt, gt_heatmap = batch_gt_data

        losses = dict()

        # pix loss
        for step, (sr, heatmap) in enumerate(zip(sr_list, heatmap_list)):
            losses[f'loss_pixel_v{step}'] = self.pixel_loss(sr, gt)
            losses[f'loss_align_v{step}'] = self.align_loss(
                heatmap, gt_heatmap)

        if self.step_counter >= self.pixel_init:
            pred = sr_list[-1]

            # perceptual loss
            if self.feature_loss:
                loss_feature = self.feature_loss(pred, gt)
                losses['loss_feature'] = loss_feature

            # gan loss for generator
            if self.gan_loss and self.discriminator:
                fake_g_pred = self.discriminator(pred)
                losses['loss_gan'] = self.gan_loss(
                    fake_g_pred, target_is_real=True, is_disc=False)

        return losses

    def train_step(self, data: List[dict],
                   optim_wrapper: OptimWrapperDict) -> Dict[str, torch.Tensor]:
        """Train step of GAN-based method.

        Args:
            data (List[dict]): Data sampled from dataloader.
            optim_wrapper (OptimWrapper): OptimWrapper instance
                used to update model parameters.

        Returns:
            Dict[str, torch.Tensor]: A ``dict`` of tensor for logging.
        """

        g_optim_wrapper = optim_wrapper['generator']

        data = self.data_preprocessor(data, True)
        batch_inputs = data['inputs']
        data_samples = data['data_samples']
        batch_gt_data = self.extract_gt_data(data_samples)

        log_vars = dict()

        with g_optim_wrapper.optim_context(self):
            batch_outputs = self.forward_train(batch_inputs, data_samples)

        if self.if_run_g():
            set_requires_grad(self.discriminator, False)

            log_vars_d = self.g_step_with_optim(
                batch_outputs=batch_outputs,
                batch_gt_data=batch_gt_data,
                optim_wrapper=optim_wrapper)

            log_vars.update(log_vars_d)

        if self.if_run_d():
            set_requires_grad(self.discriminator, True)

            sr_list, _ = batch_outputs
            gt, _ = batch_gt_data

            for _ in range(self.disc_repeat):
                # detach before function call to resolve PyTorch2.0 compile bug
                log_vars_d = self.d_step_with_optim(
                    batch_outputs=sr_list[-1].detach(),
                    batch_gt_data=gt,
                    optim_wrapper=optim_wrapper)

            log_vars.update(log_vars_d)

        if 'loss' in log_vars:
            log_vars.pop('loss')

        self.step_counter += 1

        return log_vars

    @staticmethod
    def extract_gt_data(data_samples):
        """extract gt data from data samples.

        Args:
            data_samples (list): List of DataSample.

        Returns:
            Tensor: Extract gt data.
        """

        batch_gt_img = data_samples.gt_img
        batch_gt_heatmap = data_samples.gt_heatmap

        return [batch_gt_img, batch_gt_heatmap]



# Copyright (c) OpenMMLab. All rights reserved.
import math
from functools import partial

import torch
import torch.nn as nn

from mmagic.registry import MODELS

# Note: This model is copied from Disco-Diffusion colab.
# SourceCode: https://colab.research.google.com/drive/1uGKaBOEACeinAA7jX1_zSFtj_ZW-huHS#scrollTo=XIqUfrmvLIhg # noqa


def append_dims(x, n):
    """Append dims."""
    return x[(Ellipsis, *(None, ) * (n - x.ndim))]


def expand_to_planes(x, shape):
    """Expand tensor to planes."""
    return append_dims(x, len(shape)).repeat([1, 1, *shape[2:]])


def alpha_sigma_to_t(alpha, sigma):
    """convert alpha&sigma to timestep."""
    return torch.atan2(sigma, alpha) * 2 / math.pi


def t_to_alpha_sigma(t):
    """convert timestep to alpha and sigma."""
    return torch.cos(t * math.pi / 2), torch.sin(t * math.pi / 2)


class ConvBlock(nn.Sequential):
    """Convolution Block.

    Args:
        c_in (int): Input channels.
        c_out (int): Output channels.
    """

    def __init__(self, c_in, c_out):
        super().__init__(
            nn.Conv2d(c_in, c_out, 3, padding=1),
            nn.ReLU(inplace=True),
        )


class SkipBlock(nn.Module):
    """Skip block wrapper. Wrapping main block and skip block and concat their
    outputs together.

    Args:
        main (list): A list of main modules.
        skip (nn.Module): Skip Module. If not given,
            set to ``nn.Identity()``. Defaults to None.
    """

    def __init__(self, main, skip=None):
        super().__init__()
        self.main = nn.Sequential(*main)
        self.skip = skip if skip else nn.Identity()

    def forward(self, input):
        """Forward function."""
        return torch.cat([self.main(input), self.skip(input)], dim=1)


class FourierFeatures(nn.Module):
    """Fourier features mapping MLP.

    Args:
        in_features (int): Input channels.
        out_features (int): Output channels.
        std (float): Standard deviation. Defaults to 1..
    """

    def __init__(self, in_features, out_features, std=1.):
        super().__init__()
        assert out_features % 2 == 0
        self.weight = nn.Parameter(
            torch.randn([out_features // 2, in_features]) * std)

    def forward(self, input):
        """Forward function."""
        f = 2 * math.pi * input @ self.weight.T
        return torch.cat([f.cos(), f.sin()], dim=-1)


@MODELS.register_module()
class SecondaryDiffusionImageNet2(nn.Module):
    """A smaller secondary diffusion model trained by Katherine Crowson to
    remove noise from intermediate timesteps to prepare them for CLIP.

    Ref: https://twitter.com/rivershavewings/status/1462859669454536711 # noqa
    """

    def __init__(self):
        super().__init__()
        self.in_channels = 3
        c = 64  # The base channel count
        cs = [c, c * 2, c * 2, c * 4, c * 4, c * 8]

        self.timestep_embed = FourierFeatures(1, 16)
        self.down = nn.AvgPool2d(2)
        self.up = nn.Upsample(
            scale_factor=2, mode='bilinear', align_corners=False)

        self.net = nn.Sequential(
            ConvBlock(3 + 16, cs[0]),
            ConvBlock(cs[0], cs[0]),
            SkipBlock([
                self.down,
                ConvBlock(cs[0], cs[1]),
                ConvBlock(cs[1], cs[1]),
                SkipBlock([
                    self.down,
                    ConvBlock(cs[1], cs[2]),
                    ConvBlock(cs[2], cs[2]),
                    SkipBlock([
                        self.down,
                        ConvBlock(cs[2], cs[3]),
                        ConvBlock(cs[3], cs[3]),
                        SkipBlock([
                            self.down,
                            ConvBlock(cs[3], cs[4]),
                            ConvBlock(cs[4], cs[4]),
                            SkipBlock([
                                self.down,
                                ConvBlock(cs[4], cs[5]),
                                ConvBlock(cs[5], cs[5]),
                                ConvBlock(cs[5], cs[5]),
                                ConvBlock(cs[5], cs[4]),
                                self.up,
                            ]),
                            ConvBlock(cs[4] * 2, cs[4]),
                            ConvBlock(cs[4], cs[3]),
                            self.up,
                        ]),
                        ConvBlock(cs[3] * 2, cs[3]),
                        ConvBlock(cs[3], cs[2]),
                        self.up,
                    ]),
                    ConvBlock(cs[2] * 2, cs[2]),
                    ConvBlock(cs[2], cs[1]),
                    self.up,
                ]),
                ConvBlock(cs[1] * 2, cs[1]),
                ConvBlock(cs[1], cs[0]),
                self.up,
            ]),
            ConvBlock(cs[0] * 2, cs[0]),
            nn.Conv2d(cs[0], 3, 3, padding=1),
        )

    def forward(self, input, t):
        """Forward function."""
        timestep_embed = expand_to_planes(
            self.timestep_embed(t[:, None]), input.shape)
        v = self.net(torch.cat([input, timestep_embed], dim=1))
        alphas, sigmas = map(
            partial(append_dims, n=v.ndim), t_to_alpha_sigma(t))
        pred = input * alphas - v * sigmas
        eps = input * sigmas + v * alphas
        return dict(v=v, pred=pred, eps=eps)



# Copyright (c) OpenMMLab. All rights reserved.
import torch.nn as nn
from torch import Tensor

from ..utils import default_init_weights

# def default_init_weights(module, scale=1):
#     """Initialize network weights.

#     Args:
#         modules (nn.Module): Modules to be initialized.
#         scale (float): Scale initialized weights, especially for residual
#             blocks. Default: 1.
#     """
#     for m in module.modules():
#         if isinstance(m, nn.Conv2d):
#             kaiming_init(m, a=0, mode='fan_in', bias=0)
#             m.weight.data *= scale
#         elif isinstance(m, nn.Linear):
#             kaiming_init(m, a=0, mode='fan_in', bias=0)
#             m.weight.data *= scale
#         elif isinstance(m, _BatchNorm):
#             constant_init(m.weight, val=1, bias=0)

# def make_layer(block, num_blocks, **kwarg):
#     """Make layers by stacking the same blocks.

#     Args:
#         block (nn.module): nn.module class for basic block.
#         num_blocks (int): number of blocks.

#     Returns:
#         nn.Sequential: Stacked blocks in nn.Sequential.
#     """
#     layers = []
#     for _ in range(num_blocks):
#         layers.append(block(**kwarg))
#     return nn.Sequential(*layers)


class ResidualBlockNoBN(nn.Module):
    """Residual block without BN.

    It has a style of:

    ::

        ---Conv-ReLU-Conv-+-
         |________________|

    Args:
        mid_channels (int): Channel number of intermediate features.
            Default: 64.
        res_scale (float): Used to scale the residual before addition.
            Default: 1.0.
    """

    def __init__(self, mid_channels: int = 64, res_scale: float = 1.0):
        super().__init__()
        self.res_scale = res_scale
        self.conv1 = nn.Conv2d(mid_channels, mid_channels, 3, 1, 1, bias=True)
        self.conv2 = nn.Conv2d(mid_channels, mid_channels, 3, 1, 1, bias=True)

        self.relu = nn.ReLU(inplace=True)

        # if res_scale < 1.0, use the default initialization, as in EDSR.
        # if res_scale = 1.0, use scaled kaiming_init, as in MSRResNet.
        if res_scale == 1.0:
            self.init_weights()

    def init_weights(self) -> None:
        """Initialize weights for ResidualBlockNoBN.

        Initialization methods like `kaiming_init` are for VGG-style modules.
        For modules with residual paths, using smaller std is better for
        stability and performance. We empirically use 0.1. See more details in
        "ESRGAN: Enhanced Super-Resolution Generative Adversarial Networks"
        """

        for m in [self.conv1, self.conv2]:
            default_init_weights(m, 0.1)

    def forward(self, x: Tensor) -> Tensor:
        """Forward function.

        Args:
            x (Tensor): Input tensor with shape (n, c, h, w).

        Returns:
            Tensor: Forward results.
        """

        identity = x
        out = self.conv2(self.relu(self.conv1(x)))
        return identity + out * self.res_scale



import torch
from mmengine.model import BaseModule
from torch import nn

from mmagic.registry import MODELS


@MODELS.register_module()
class RDNNet(BaseModule):
    """RDN model for single image super-resolution.

    Paper: Residual Dense Network for Image Super-Resolution

    Adapted from 'https://github.com/yjn870/RDN-pytorch.git'
    'RDN-pytorch/blob/master/models.py'
    Copyright (c) 2021, JaeYun Yeo, under MIT License.

    Most of the implementation follows the implementation in:
    'https://github.com/sanghyun-son/EDSR-PyTorch.git'
    'EDSR-PyTorch/blob/master/src/model/rdn.py'
    Copyright (c) 2017, sanghyun-son, under MIT license.

    Args:
        in_channels (int): Channel number of inputs.
        out_channels (int): Channel number of outputs.
        mid_channels (int): Channel number of intermediate features.
            Default: 64.
        num_blocks (int): Block number in the trunk network. Default: 16.
        upscale_factor (int): Upsampling factor. Support 2^n and 3.
            Default: 4.
        num_layer (int): Layer number in the Residual Dense Block.
            Default: 8.
        channel_growth(int): Channels growth in each layer of RDB.
            Default: 64.
    """

    def __init__(self,
                 in_channels,
                 out_channels,
                 mid_channels=64,
                 num_blocks=16,
                 upscale_factor=4,
                 num_layers=8,
                 channel_growth=64):

        super().__init__()
        self.mid_channels = mid_channels
        self.channel_growth = channel_growth
        self.num_blocks = num_blocks
        self.num_layers = num_layers

        # shallow feature extraction
        self.sfe1 = nn.Conv2d(
            in_channels, mid_channels, kernel_size=3, padding=3 // 2)
        self.sfe2 = nn.Conv2d(
            mid_channels, mid_channels, kernel_size=3, padding=3 // 2)

        # residual dense blocks
        self.rdbs = nn.ModuleList()
        for _ in range(self.num_blocks):
            self.rdbs.append(
                RDB(self.mid_channels, self.channel_growth, self.num_layers))

        # global feature fusion
        self.gff = nn.Sequential(
            nn.Conv2d(
                self.mid_channels * self.num_blocks,
                self.mid_channels,
                kernel_size=1),
            nn.Conv2d(
                self.mid_channels,
                self.mid_channels,
                kernel_size=3,
                padding=3 // 2))

        # up-sampling
        assert 2 <= upscale_factor <= 4
        if upscale_factor == 2 or upscale_factor == 4:
            self.upscale = []
            for _ in range(upscale_factor // 2):
                self.upscale.extend([
                    nn.Conv2d(
                        self.mid_channels,
                        self.mid_channels * (2**2),
                        kernel_size=3,
                        padding=3 // 2),
                    nn.PixelShuffle(2)
                ])
            self.upscale = nn.Sequential(*self.upscale)
        else:
            self.upscale = nn.Sequential(
                nn.Conv2d(
                    self.mid_channels,
                    self.mid_channels * (upscale_factor**2),
                    kernel_size=3,
                    padding=3 // 2), nn.PixelShuffle(upscale_factor))

        self.output = nn.Conv2d(
            self.mid_channels, out_channels, kernel_size=3, padding=3 // 2)

    def forward(self, x):
        """Forward function.

        Args:
            x (Tensor): Input tensor with shape (n, c, h, w).

        Returns:
            Tensor: Forward results.
        """

        sfe1 = self.sfe1(x)
        sfe2 = self.sfe2(sfe1)

        x = sfe2
        local_features = []
        for i in range(self.num_blocks):
            x = self.rdbs[i](x)
            local_features.append(x)

        x = self.gff(torch.cat(local_features, 1)) + sfe1
        # global residual learning
        x = self.upscale(x)
        x = self.output(x)
        return x


class DenseLayer(BaseModule):
    """Dense layer.

    Args:
        in_channels (int): Channel number of inputs.
        out_channels (int): Channel number of outputs.
    """

    def __init__(self, in_channels, out_channels):
        super().__init__()
        self.conv = nn.Conv2d(
            in_channels, out_channels, kernel_size=3, padding=3 // 2)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        """Forward function.

        Args:
            x (Tensor): Input tensor with shape (n, c_in, h, w).

        Returns:
            Tensor: Forward results, tensor with shape (n, c_in+c_out, h, w).
        """
        return torch.cat([x, self.relu(self.conv(x))], 1)


class RDB(BaseModule):
    """Residual Dense Block of Residual Dense Network.

    Args:
        in_channels (int): Channel number of inputs.
        channel_growth (int): Channels growth in each layer.
        num_layers (int): Layer number in the Residual Dense Block.
    """

    def __init__(self, in_channels, channel_growth, num_layers):
        super().__init__()
        self.layers = nn.Sequential(*[
            DenseLayer(in_channels + channel_growth * i, channel_growth)
            for i in range(num_layers)
        ])

        # local feature fusion
        self.lff = nn.Conv2d(
            in_channels + channel_growth * num_layers,
            in_channels,
            kernel_size=1)

    def forward(self, x):
        """Forward function.

        Args:
            x (Tensor): Input tensor with shape (n, c, h, w).

        Returns:
            Tensor: Forward results.
        """
        return x + self.lff(self.layers(x))  # local residual learning



_base_ = ['swinir_x3s48w8d6e180_8xb4-lr2e-4-500k_div2k.py']

experiment_name = 'swinir_x3s64w8d6e180_8xb4-lr2e-4-500k_df2k'
work_dir = f'./work_dirs/{experiment_name}'
save_dir = './work_dirs/'

scale = 3
img_size = 64

# model settings
model = dict(generator=dict(img_size=img_size))

# modify patch size of train_pipeline
train_pipeline = _base_.train_pipeline
train_pipeline[3]['gt_patch_size'] = img_size * scale

# dataset settings
dataset_type = 'BasicImageDataset'
data_root = 'data'

train_dataloader = dict(
    num_workers=4,
    batch_size=4,
    drop_last=True,
    persistent_workers=False,
    sampler=dict(type='InfiniteSampler', shuffle=True),
    dataset=dict(
        type=dataset_type,
        ann_file='meta_info_DF2K3450sub_GT.txt',
        metainfo=dict(dataset_type='div2k', task_name='sisr'),
        data_root=data_root + '/DF2K',
        data_prefix=dict(
            img='DF2K_train_LR_bicubic/X3_sub', gt='DF2K_train_HR_sub'),
        filename_tmpl=dict(img='{}', gt='{}'),
        pipeline=train_pipeline))



scale_test_list = [2, 3, 4, 6, 18, 30]

test_pipelines = [[
    dict(
        type='LoadImageFromFile',
        key='gt',
        color_type='color',
        channel_order='rgb',
        imdecode_backend='cv2'),
    dict(
        type='RandomDownSampling', scale_min=scale_test, scale_max=scale_test),
    dict(type='GenerateCoordinateAndCell', scale=scale_test, reshape_gt=False),
    dict(type='PackInputs')
] for scale_test in scale_test_list]

# test config for Set5
set5_dataloaders = [
    dict(
        num_workers=4,
        persistent_workers=False,
        drop_last=False,
        sampler=dict(type='DefaultSampler', shuffle=False),
        dataset=dict(
            type='BasicImageDataset',
            metainfo=dict(dataset_type='set5', task_name='sisr'),
            data_root='data/Set5',
            data_prefix=dict(img='LRbicx4', gt='GTmod12'),
            pipeline=test_pipeline)) for test_pipeline in test_pipelines
]
set5_evaluators = [[
    dict(type='PSNR', crop_border=scale, prefix=f'Set5x{scale}'),
    dict(type='SSIM', crop_border=scale, prefix=f'Set5x{scale}'),
] for scale in scale_test_list]

# test config for Set14
set14_dataloaders = [
    dict(
        num_workers=4,
        persistent_workers=False,
        drop_last=False,
        sampler=dict(type='DefaultSampler', shuffle=False),
        dataset=dict(
            type='BasicImageDataset',
            metainfo=dict(dataset_type='set14', task_name='sisr'),
            data_root='data/Set14',
            data_prefix=dict(img='LRbicx4', gt='GTmod12'),
            pipeline=test_pipeline)) for test_pipeline in test_pipelines
]
set14_evaluators = [[
    dict(type='PSNR', crop_border=scale, prefix=f'Set14x{scale}'),
    dict(type='SSIM', crop_border=scale, prefix=f'Set14x{scale}'),
] for scale in scale_test_list]

# test config for DIV2K
div2k_dataloaders = [
    dict(
        num_workers=4,
        persistent_workers=False,
        drop_last=False,
        sampler=dict(type='DefaultSampler', shuffle=False),
        dataset=dict(
            type='BasicImageDataset',
            ann_file='meta_info_DIV2K100sub_GT.txt',
            metainfo=dict(dataset_type='div2k', task_name='sisr'),
            data_root='data/DIV2K',
            data_prefix=dict(
                img='DIV2K_train_LR_bicubic/X4_sub', gt='DIV2K_train_HR_sub'),
            pipeline=test_pipeline)) for test_pipeline in test_pipelines
]
div2k_evaluators = [[
    dict(type='PSNR', crop_border=scale, prefix=f'DIV2Kx{scale}'),
    dict(type='SSIM', crop_border=scale, prefix=f'DIV2Kx{scale}'),
] for scale in scale_test_list]

# test config
test_cfg = dict(type='MultiTestLoop')
test_dataloader = [
    *set5_dataloaders,
    *set14_dataloaders,
    *div2k_dataloaders,
]
test_evaluator = [
    *set5_evaluators,
    *set14_evaluators,
    *div2k_evaluators,
]



# Copyright (c) OpenMMLab. All rights reserved.
from mmcv.utils import collect_env as collect_base_env
from mmengine.utils import get_git_hash

import mmagic


def collect_env():
    """Collect the information of the running environments."""
    env_info = collect_base_env()
    env_info['MMagic'] = f'{mmagic.__version__}+{get_git_hash()[:7]}'

    return env_info


if __name__ == '__main__':
    for name, val in collect_env().items():
        print('{}: {}'.format(name, val))



# Copyright (c) OpenMMLab. All rights reserved.
import torch.nn as nn
from mmengine.model import BaseModule

from mmagic.registry import MODELS


@MODELS.register_module()
class TTSRDiscriminator(BaseModule):
    """A discriminator for TTSR.

    Args:
        in_channels (int): Channel number of inputs. Default: 3.
        in_size (int): Size of input image. Default: 160.
    """

    def __init__(self, in_channels=3, in_size=160):
        super().__init__()

        self.body = nn.Sequential(
            nn.Conv2d(in_channels, 32, 3, 1, 1), nn.LeakyReLU(0.2),
            nn.Conv2d(32, 32, 3, 2, 1), nn.LeakyReLU(0.2),
            nn.Conv2d(32, 64, 3, 1, 1), nn.LeakyReLU(0.2),
            nn.Conv2d(64, 64, 3, 2, 1), nn.LeakyReLU(0.2),
            nn.Conv2d(64, 128, 3, 1, 1), nn.LeakyReLU(0.2),
            nn.Conv2d(128, 128, 3, 2, 1), nn.LeakyReLU(0.2),
            nn.Conv2d(128, 256, 3, 1, 1), nn.LeakyReLU(0.2),
            nn.Conv2d(256, 256, 3, 2, 1), nn.LeakyReLU(0.2),
            nn.Conv2d(256, 512, 3, 1, 1), nn.LeakyReLU(0.2),
            nn.Conv2d(512, 512, 3, 2, 1), nn.LeakyReLU(0.2))

        self.last = nn.Sequential(
            nn.Linear(in_size // 32 * in_size // 32 * 512, 1024),
            nn.LeakyReLU(0.2), nn.Linear(1024, 1))

    def forward(self, x):
        """Forward function.

        Args:
            x (Tensor): Input tensor with shape (n, c, h, w).

        Returns:
            Tensor: Forward results.
        """

        x = self.body(x)
        x = x.view(x.size(0), -1)
        x = self.last(x)

        return x



test_pipeline = [
    dict(
        type='LoadImageFromFile',
        key='img',
        color_type='color',
        channel_order='rgb',
        imdecode_backend='cv2'),
    dict(
        type='LoadImageFromFile',
        key='gt',
        color_type='color',
        channel_order='rgb',
        imdecode_backend='cv2'),
    dict(type='PackInputs')
]

gopro_data_root = 'data/GoPro'
gopro_dataloader = dict(
    num_workers=4,
    persistent_workers=False,
    drop_last=False,
    sampler=dict(type='DefaultSampler', shuffle=False),
    dataset=dict(
        type='BasicImageDataset',
        metainfo=dict(dataset_type='GoPro', task_name='deblurring'),
        data_root=gopro_data_root,
        data_prefix=dict(img='input', gt='target'),
        pipeline=test_pipeline))
gopro_evaluator = [
    dict(type='PSNR', prefix='GoPro'),
    dict(type='SSIM', prefix='GoPro'),
]

hide_data_root = 'data/HIDE'
hide_dataloader = dict(
    num_workers=4,
    persistent_workers=False,
    drop_last=False,
    sampler=dict(type='DefaultSampler', shuffle=False),
    dataset=dict(
        type='BasicImageDataset',
        metainfo=dict(dataset_type='HIDE', task_name='deblurring'),
        data_root=hide_data_root,
        data_prefix=dict(img='input', gt='target'),
        pipeline=test_pipeline))
hide_evaluator = [
    dict(type='PSNR', prefix='HIDE'),
    dict(type='SSIM', prefix='HIDE'),
]

realblurj_data_root = 'data/RealBlur_J'
realblurj_dataloader = dict(
    num_workers=4,
    persistent_workers=False,
    drop_last=False,
    sampler=dict(type='DefaultSampler', shuffle=False),
    dataset=dict(
        type='BasicImageDataset',
        metainfo=dict(dataset_type='RealBlur_J', task_name='deblurring'),
        data_root=realblurj_data_root,
        data_prefix=dict(img='input', gt='target'),
        pipeline=test_pipeline))
realblurj_evaluator = [
    dict(type='PSNR', convert_to='Y', prefix='RealBlurJ'),
    dict(type='SSIM', convert_to='Y', prefix='RealBlurJ'),
]

realblurr_data_root = 'data/RealBlur_R'
realblurr_dataloader = dict(
    num_workers=4,
    persistent_workers=False,
    drop_last=False,
    sampler=dict(type='DefaultSampler', shuffle=False),
    dataset=dict(
        type='BasicImageDataset',
        metainfo=dict(dataset_type='RealBlur_R', task_name='deblurring'),
        data_root=realblurr_data_root,
        data_prefix=dict(img='input', gt='target'),
        pipeline=test_pipeline))
realblurr_evaluator = [
    dict(type='PSNR', convert_to='Y', prefix='RealBlurR'),
    dict(type='SSIM', convert_to='Y', prefix='RealBlurR'),
]

# test config
test_cfg = dict(type='MultiTestLoop')
test_dataloader = [
    gopro_dataloader,
    hide_dataloader,
    realblurj_dataloader,
    realblurr_dataloader,
]
test_evaluator = [
    gopro_evaluator,
    hide_evaluator,
    realblurj_evaluator,
    realblurr_evaluator,
]



_base_ = [
    'restormer_official_dpdd-single.py',
]

experiment_name = 'restormer_official_dpdd-dual'
work_dir = f'./work_dirs/{experiment_name}'
save_dir = './work_dirs/'

# modify PackInputs
test_dataloader = _base_.test_dataloader
for dataloader in test_dataloader:
    test_pipeline = dataloader['dataset']['pipeline']
    test_pipeline[4] = dict(type='PackInputs', keys=['imgL', 'imgR'])

# model settings
model = dict(
    generator=dict(inp_channels=6, dual_pixel_task=True),
    data_preprocessor=dict(type='DataPreprocessor'))



# Copyright (c) OpenMMLab. All rights reserved.
import os
from logging import WARNING
from typing import Any, List, Optional, Union

import torch
from mmengine import print_log
from mmengine.model import BaseModule
from torch import dtype as TORCH_DTYPE

dtype_mapping = {
    'float32': torch.float32,
    'float16': torch.float16,
    'fp32': torch.float32,
    'fp16': torch.float16,
    'half': torch.float16,
}


class DiffusersWrapper(BaseModule):
    """Wrapper for models from HuggingFace Diffusers. This wrapper will be set
    a attribute called `_module_cls` by wrapping function and will be used to
    initialize the model structure.

    Example:
    >>> 1. Load pretrained model from HuggingFace Space.
    >>> config = dict(
    >>>     type='ControlNetModel',  # has been registered in `MODELS`
    >>>     from_pretrained='lllyasviel/sd-controlnet-canny',
    >>>     torch_dtype=torch.float16)
    >>> controlnet = MODELS.build(config)

    >>> 2. Initialize model with pre-defined configs.
    >>> config = dict(
    >>>     type='ControlNetModel',  # has been registered in `MODELS`
    >>>     from_config='lllyasviel/sd-controlnet-canny',
    >>>     cache_dir='~/.cache/OpenMMLab')
    >>> controlnet = MODELS.build(config)

    >>> 3. Initialize model with own defined arguments
    >>> config = dict(
    >>>     type='ControlNetModel',  # has been registered in `MODELS`
    >>>     in_channels=3,
    >>>     down_block_types=['DownBlock2D'],
    >>>     block_out_channels=(32, ),
    >>>     conditioning_embedding_out_channels=(16, ))
    >>> controlnet = MODELS.build(config)

    Args:
        from_pretrained (Union[str, os.PathLike], optional): The *model id*
            of a pretrained model or a path to a *directory* containing
            model weights and config. Please refers to
            `diffusers.model.modeling_utils.ModelMixin.from_pretrained`
            for more detail. Defaults to None.
        from_config (Union[str, os.PathLike], optional): The *model id*
            of a pretrained model or a path to a *directory* containing
            model weights and config. Please refers to
            `diffusers.configuration_utils.ConfigMixin.load_config`
            for more detail. Defaults to None.
        init_cfg (dict or List[dict], optional): Initialization config dict.
            Noted that, in `DiffuserWrapper`, if you want to load pretrained
            weight from HuggingFace space, please use `from_pretrained`
            argument instead of using `init_cfg`. Defaults to None.

        *args, **kwargs: If `from_pretrained` is passed, *args and **kwargs
            will be passed to `from_pretrained` function. If `from_config`
            is passed, *args and **kwargs will be passed to `load_config`
            function.  Otherwise, *args and **kwargs will be used to
            initialize the model by `self._module_cls(*args, **kwargs)`.
    """

    def __init__(self,
                 from_pretrained: Optional[Union[str, os.PathLike]] = None,
                 from_config: Optional[Union[str, os.PathLike]] = None,
                 dtype: Optional[Union[str, TORCH_DTYPE]] = None,
                 init_cfg: Union[dict, List[dict], None] = None,
                 *args,
                 **kwargs):
        super().__init__(init_cfg)

        module_cls = self._module_cls
        assert not (from_pretrained and from_config), (
            '\'from_pretrained\' and \'from_config\' should not be passed '
            'at the same time.')

        self._from_pretrained = from_pretrained
        self._from_config = from_config

        if from_pretrained is not None:
            self.model = module_cls.from_pretrained(from_pretrained, *args,
                                                    **kwargs)
            # weight has been initialized from pretrained, therefore we
            # `self._is_init` as True manually
            self._is_init = True
        elif from_config is not None:
            _config = module_cls.load_config(from_config, *args, **kwargs)
            self.model = module_cls(**_config)
        else:
            self.model = module_cls(*args, **kwargs)

        if dtype is not None:
            if isinstance(dtype, str):
                assert dtype in dtype_mapping, (
                    'Only support following dtype string: '
                    f'{list(dtype_mapping.keys())}, but receive {dtype}.')
                dtype = dtype_mapping[dtype]
            self.model.to(dtype)
            print_log(f'Set model dtype to \'{dtype}\'.', 'current')

        self.config = self.model.config

    def init_weights(self):
        """Initialize the weights.

        If type is 'Pretrained' but the model has be loaded from `repo_id`, a
        warning will be raised.
        """
        if self.init_cfg and self.init_cfg['type'] == 'Pretrained':
            if self._from_pretrained is not None:
                print_log(
                    'Has been loaded from pretrained model from '
                    f'\'{self._from_pretrained}\'. Your behavior is '
                    'very dangerous.', 'current', WARNING)
        super().init_weights()

    def __getattr__(self, name: str) -> Any:
        """This function provide a way to access the attributes of the wrapped
        model.

        Args:
            name (str): The name of the attribute.

        Returns:
            Any: The got attribute.
        """
        # Q: why we need end of recursion for 'model'?
        # A: In `nn.Module.__setattr__`, if value is instance of `nn.Module`,
        #   it will be removed from `__dict__` and added to
        #   `__dict__._modules`. Therefore, `model` cannot be found in
        #   `self.__dict__`. When we call `self.model`, python cannot found
        #   'model' in `self.__dict__` and then `self.__getattr__('model')`
        #   will be called. If we call `self.model` in `self.__getattr__`
        #   which does not have any exit about 'model',`RecursionError`
        #   will be raised.
        if name == 'model':
            return super().__getattr__('model')

        try:
            return getattr(self.model, name)
        except AttributeError:
            try:
                return super().__getattr__(name)
            except AttributeError:
                raise AttributeError('\'name\' cannot be found in both '
                                     f'\'{self.__class__.__name__}\' and '
                                     f'\'{self.__class__.__name__}.model\'.')

    def __repr__(self):
        """The representation of the wrapper."""
        s = super().__repr__()
        prefix = f'Wrapped Module Class: {self._module_cls}\n'
        prefix += f'Wrapped Module Name: {self._module_name}\n'
        if self._from_pretrained:
            prefix += f'From Pretrained: {self._from_pretrained}\n'
        if self._from_config:
            prefix += f'From Config: {self._from_config}\n'
        s = prefix + s
        return s

    def forward(self, *args, **kwargs) -> Any:
        """Forward function of wrapped module.

        Args:
            *args, **kwargs: The arguments of the wrapped module.

        Returns:
            Any: The output of wrapped module's forward function.
        """
        return self.model(*args, **kwargs)



# Copyright (c) OpenMMLab. All rights reserved.
import torch.nn as nn
from mmengine.model import BaseModule

from mmagic.registry import MODELS


@MODELS.register_module()
class ModifiedVGG(BaseModule):
    """A modified VGG discriminator with input size 128 x 128.

    It is used to train SRGAN and ESRGAN.

    Args:
        in_channels (int): Channel number of inputs. Default: 3.
        mid_channels (int): Channel number of base intermediate features.
            Default: 64.
    """

    def __init__(self, in_channels, mid_channels):
        super().__init__()

        self.conv0_0 = nn.Conv2d(in_channels, mid_channels, 3, 1, 1, bias=True)
        self.conv0_1 = nn.Conv2d(
            mid_channels, mid_channels, 4, 2, 1, bias=False)
        self.bn0_1 = nn.BatchNorm2d(mid_channels, affine=True)

        self.conv1_0 = nn.Conv2d(
            mid_channels, mid_channels * 2, 3, 1, 1, bias=False)
        self.bn1_0 = nn.BatchNorm2d(mid_channels * 2, affine=True)
        self.conv1_1 = nn.Conv2d(
            mid_channels * 2, mid_channels * 2, 4, 2, 1, bias=False)
        self.bn1_1 = nn.BatchNorm2d(mid_channels * 2, affine=True)

        self.conv2_0 = nn.Conv2d(
            mid_channels * 2, mid_channels * 4, 3, 1, 1, bias=False)
        self.bn2_0 = nn.BatchNorm2d(mid_channels * 4, affine=True)
        self.conv2_1 = nn.Conv2d(
            mid_channels * 4, mid_channels * 4, 4, 2, 1, bias=False)
        self.bn2_1 = nn.BatchNorm2d(mid_channels * 4, affine=True)

        self.conv3_0 = nn.Conv2d(
            mid_channels * 4, mid_channels * 8, 3, 1, 1, bias=False)
        self.bn3_0 = nn.BatchNorm2d(mid_channels * 8, affine=True)
        self.conv3_1 = nn.Conv2d(
            mid_channels * 8, mid_channels * 8, 4, 2, 1, bias=False)
        self.bn3_1 = nn.BatchNorm2d(mid_channels * 8, affine=True)

        self.conv4_0 = nn.Conv2d(
            mid_channels * 8, mid_channels * 8, 3, 1, 1, bias=False)
        self.bn4_0 = nn.BatchNorm2d(mid_channels * 8, affine=True)
        self.conv4_1 = nn.Conv2d(
            mid_channels * 8, mid_channels * 8, 4, 2, 1, bias=False)
        self.bn4_1 = nn.BatchNorm2d(mid_channels * 8, affine=True)

        self.linear1 = nn.Linear(mid_channels * 8 * 4 * 4, 100)
        self.linear2 = nn.Linear(100, 1)

        # activation function
        self.lrelu = nn.LeakyReLU(negative_slope=0.2, inplace=True)

    def forward(self, x):
        """Forward function.

        Args:
            x (Tensor): Input tensor with shape (n, c, h, w).

        Returns:
            Tensor: Forward results.
        """

        assert x.size(2) == 128 and x.size(3) == 128, (
            f'Input spatial size must be 128x128, '
            f'but received {x.size()}.')

        feat = self.lrelu(self.conv0_0(x))
        feat = self.lrelu(self.bn0_1(
            self.conv0_1(feat)))  # output spatial size: (64, 64)

        feat = self.lrelu(self.bn1_0(self.conv1_0(feat)))
        feat = self.lrelu(self.bn1_1(
            self.conv1_1(feat)))  # output spatial size: (32, 32)

        feat = self.lrelu(self.bn2_0(self.conv2_0(feat)))
        feat = self.lrelu(self.bn2_1(
            self.conv2_1(feat)))  # output spatial size: (16, 16)

        feat = self.lrelu(self.bn3_0(self.conv3_0(feat)))
        feat = self.lrelu(self.bn3_1(
            self.conv3_1(feat)))  # output spatial size: (8, 8)

        feat = self.lrelu(self.bn4_0(self.conv4_0(feat)))
        feat = self.lrelu(self.bn4_1(
            self.conv4_1(feat)))  # output spatial size: (4, 4)

        feat = feat.view(feat.size(0), -1)
        feat = self.lrelu(self.linear1(feat))
        out = self.linear2(feat)
        return out



# Copyright (c) OpenMMLab. All rights reserved.
from .clip_wrapper import ClipWrapper
from .disco import DiscoDiffusion
from .guider import ImageTextGuider
from .secondary_model import SecondaryDiffusionImageNet2, alpha_sigma_to_t

__all__ = [
    'DiscoDiffusion', 'ImageTextGuider', 'ClipWrapper',
    'SecondaryDiffusionImageNet2', 'alpha_sigma_to_t'
]



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, Tuple

import torch
from mmengine.optim import OptimWrapper
from torch import Tensor

from mmagic.models.base_models import BaseGAN
from mmagic.models.losses import gradient_penalty_loss
from mmagic.registry import MODELS
from mmagic.structures import DataSample


@MODELS.register_module()
class WGANGP(BaseGAN):
    """Implementation of `Improved Training of Wasserstein GANs`.

    Paper link: https://arxiv.org/pdf/1704.00028

    Detailed architecture can be found in
    :class:`~mmagic.models.editors.wgan_gp.WGANGPGenerator`
    and
    :class:`~mmagic.models.editors.wgan_gp.WGANGPDiscriminator`
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # gradient penalty loss settings
        self.gp_norm_mode = self.loss_config.get('norm_mode', 'HWC')
        self.gp_loss_weight = self.loss_config.get('loss_weight', 10)

    def disc_loss(self, real_data: Tensor, fake_data: Tensor,
                  disc_pred_fake: Tensor, disc_pred_real: Tensor) -> Tuple:
        r"""Get disc loss. WGAN-GP use the wgan loss and gradient penalty to
        train the discriminator.

        Args:
            real_data (Tensor): Real input data.
            fake_data (Tensor): Fake input data.
            disc_pred_fake (Tensor): Discriminator's prediction of the fake
                images.
            disc_pred_real (Tensor): Discriminator's prediction of the real
                images.

        Returns:
            tuple[Tensor, dict]: Loss value and a dict of log variables.
        """
        losses_dict = dict()
        losses_dict['loss_disc_fake'] = disc_pred_fake.mean()
        losses_dict['loss_disc_real'] = -disc_pred_real.mean()

        # Gradient Penalty loss
        losses_dict['loss_gp'] = self.gp_loss_weight * gradient_penalty_loss(
            self.discriminator,
            real_data,
            fake_data,
            norm_mode=self.gp_norm_mode)
        loss, log_var = self.parse_losses(losses_dict)
        return loss, log_var

    def gen_loss(self, disc_pred_fake: Tensor) -> Tuple:
        """Get gen loss. DCGAN use the wgan loss to train the generator.

        Args:
            disc_pred_fake (Tensor): Discriminator's prediction of the fake
                images.

        Returns:
            tuple[Tensor, dict]: Loss value and a dict of log variables.
        """
        losses_dict = dict()
        losses_dict['loss_gen'] = -disc_pred_fake.mean()
        loss, log_var = self.parse_losses(losses_dict)
        return loss, log_var

    def train_discriminator(self, inputs: dict, data_samples: DataSample,
                            optimizer_wrapper: OptimWrapper
                            ) -> Dict[str, Tensor]:
        """Train discriminator.

        Args:
            inputs (dict): Inputs from dataloader.
            data_samples (DataSample): Data samples from dataloader.
            optim_wrapper (OptimWrapper): OptimWrapper instance used to update
                model parameters.
        Returns:
            Dict[str, Tensor]: A ``dict`` of tensor for logging.
        """
        real_imgs = data_samples.gt_img

        num_batches = real_imgs.shape[0]

        noise_batch = self.noise_fn(num_batches=num_batches)
        with torch.no_grad():
            fake_imgs = self.generator(noise=noise_batch, return_noise=False)

        disc_pred_fake = self.discriminator(fake_imgs)
        disc_pred_real = self.discriminator(real_imgs)

        parsed_losses, log_vars = self.disc_loss(real_imgs, fake_imgs,
                                                 disc_pred_fake,
                                                 disc_pred_real)
        optimizer_wrapper.update_params(parsed_losses)
        return log_vars

    def train_generator(self, inputs: dict, data_samples: DataSample,
                        optimizer_wrapper: OptimWrapper) -> Dict[str, Tensor]:
        """Train generator.

        Args:
            inputs (dict): Inputs from dataloader.
            data_samples (DataSample): Data samples from dataloader.
                Do not used in generator's training.
            optim_wrapper (OptimWrapper): OptimWrapper instance used to update
                model parameters.

        Returns:
            Dict[str, Tensor]: A ``dict`` of tensor for logging.
        """
        num_batches = len(data_samples)

        noise = self.noise_fn(num_batches=num_batches)
        fake_imgs = self.generator(noise=noise, return_noise=False)

        disc_pred_fake = self.discriminator(fake_imgs)
        parsed_loss, log_vars = self.gen_loss(disc_pred_fake)

        optimizer_wrapper.update_params(parsed_loss)
        return log_vars
